import magicattr

from dspy.predict.parallel import Parallel
from dspy.primitives.module import BaseModule
from dspy.utils.callback import with_callbacks


class ProgramMeta(type):
    pass


class Module(BaseModule, metaclass=ProgramMeta):
    def _base_init(self):
        self._compiled = False

    def __init__(self, callbacks=None):
        self.callbacks = callbacks or []
        self._compiled = False

    @with_callbacks
    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def named_predictors(self):
        from dspy.predict.predict import Predict

        return [(name, param) for name, param in self.named_parameters() if isinstance(param, Predict)]

    def predictors(self):
        return [param for _, param in self.named_predictors()]

    def set_lm(self, lm):
        for _, param in self.named_predictors():
            param.lm = lm

    def get_lm(self):
        all_used_lms = [param.lm for _, param in self.named_predictors()]

        if len(set(all_used_lms)) == 1:
            return all_used_lms[0]

        raise ValueError("Multiple LMs are being used in the module. There's no unique LM to return.")

    def __repr__(self):
        s = []

        for name, param in self.named_predictors():
            s.append(f"{name} = {param}")

        return "\n".join(s)

    def map_named_predictors(self, func):
        """Applies a function to all named predictors."""
        for name, predictor in self.named_predictors():
            set_attribute_by_name(self, name, func(predictor))
        return self

    # def activate_assertions(self, handler=backtrack_handler, **handler_args):
    #     """
    #     Activates assertions for the module.
    #     The default handler is the backtrack_handler.
    #     """
    #     assert_transform_module(self, handler, **handler_args)
    #     return self

    # def __deepcopy__(self, memo):
    #     # memo is a dict of id's to copies already made during the current call
    #     # Check if the object is already copied
    #     if id(self) in memo:
    #         return memo[id(self)]

    #     print(f"Deep copying {self.__class__.__name__}...")

    #     new_copy = copy.copy(self)
    #     memo[id(self)] = new_copy

    #     for k, v in self.__dict__.items():
    #         print(f"Copying attribute {k} of type {type(v)}...")
    #         setattr(new_copy, k, copy.deepcopy(v, memo))
    #         print("Done")

    #     return new_copy

    def batch(
        self,
        examples,
        num_threads: int = 32,
        max_errors: int = 10,
        return_failed_examples: bool = False,
        provide_traceback: bool = False,
        disable_progress_bar: bool = False,
    ):
        """
        Processes a list of dspy.Example instances in parallel using the Parallel module.

        :param examples: List of dspy.Example instances to process.
        :param batch_size: Number of threads to use for parallel processing.
        :param max_errors: Maximum number of errors allowed before stopping execution.
        :param return_failed_examples: Whether to return failed examples and exceptions.
        :param provide_traceback: Whether to include traceback information in error logs.
        :return: List of results, and optionally failed examples and exceptions.
        """
        # Create a list of execution pairs (self, example)
        exec_pairs = [(self, example.inputs()) for example in examples]

        # Create an instance of Parallel
        parallel_executor = Parallel(
            num_threads=num_threads,
            max_errors=max_errors,
            return_failed_examples=return_failed_examples,
            provide_traceback=provide_traceback,
            disable_progress_bar=disable_progress_bar,
        )

        # Execute the forward method of Parallel
        if return_failed_examples:
            results, failed_examples, exceptions = parallel_executor.forward(exec_pairs)
            return results, failed_examples, exceptions
        else:
            results = parallel_executor.forward(exec_pairs)
            return results


def set_attribute_by_name(obj, name, value):
    magicattr.set(obj, name, value)


Program = Module



import re

import dspy
from dspy.signatures.signature import ensure_signature

from ..primitives.program import Module
from ..primitives.python_interpreter import PythonInterpreter


class ProgramOfThought(Module):
    def __init__(self, signature, max_iters=3):
        super().__init__()
        self.signature = signature = ensure_signature(signature)
        self.max_iters = max_iters

        self.input_fields = signature.input_fields
        self.output_fields = signature.output_fields

        assert len(self.output_fields) == 1, "PoT only supports one output field."

        self.output_field_name = next(iter(self.output_fields))
        inputs_ = ", ".join(
            [f"`{field_name}`" for field_name in self.input_fields.keys()],
        )
        outputs_ = f"`{self.output_field_name}`"

        assert len(self.output_fields) == 1, "PoT only supports one output field."

        instr = []
        instr.append(
            f"You will be given {inputs_} and you will respond with {outputs_}.",
        )
        instr.append(
            f"Generating executable Python code that programmatically computes the correct {outputs_}.",
        )
        instr.append(
            f"After you're done with the computation, make sure the last line in your code evaluates to the correct value for {outputs_}.",
        )
        instr = "\n".join(instr)

        self.code_generate = dspy.ChainOfThought(
            dspy.Signature(
                self._generate_signature("generate").fields,
                self._generate_instruction("generate"),
            ),
        )
        self.code_regenerate = dspy.ChainOfThought(
            dspy.Signature(
                self._generate_signature("regenerate").fields,
                self._generate_instruction("regenerate"),
            ),
        )
        self.generate_answer = dspy.ChainOfThought(
            dspy.Signature(
                self._generate_signature("answer").fields,
                self._generate_instruction("answer"),
            ),
        )
    def _generate_signature(self, mode):
        signature_dict = dict(self.input_fields)
        fields_for_mode = {
            "generate": {
                "generated_code": dspy.OutputField(
                    prefix="Code:",
                    desc="python code that answers the question",
                    format=str,
                ),
            },
            "regenerate": {
                "previous_code": dspy.InputField(
                    prefix="Previous Code:",
                    desc="previously-generated python code that errored",
                    format=str,
                ),
                "error": dspy.InputField(
                    prefix="Error:",
                    desc="error message from previously-generated python code",
                ),
                "generated_code": dspy.OutputField(
                    prefix="Code:",
                    desc="python code that answers the question",
                    format=str,
                ),
            },
            "answer": {
                "final_generated_code": dspy.InputField(
                    prefix="Code:",
                    desc="python code that answers the question",
                    format=str,
                ),
                "code_output": dspy.InputField(
                    prefix="Code Output:",
                    desc="output of previously-generated python code",
                ),
                self.output_field_name: self.signature.fields[self.output_field_name],
            },
        }
        signature_dict.update(fields_for_mode[mode])
        return dspy.Signature(signature_dict)

    def _generate_instruction(self, mode):
        mode_inputs = ", ".join(
            [
                f"`{field_name}`"
                for field_name in self._generate_signature(mode).input_fields
            ],
        )
        mode_outputs = f"`{self.output_field_name}`"
        if mode == "generate":
            instr = [
                f"You will be given {mode_inputs} and you will respond with {mode_outputs}.",
                f"Generating executable Python code that programmatically computes the correct {mode_outputs}.",
                f"After you're done with the computation, make sure the last line in your code evaluates to the correct value for {mode_outputs}.",
            ]
        elif mode == "regenerate":
            instr = [
                f"You are given {mode_inputs} due to an error in previous code.",
                "Your task is to correct the error and provide the new `generated_code`.",
            ]
        else:  # mode == 'answer'
            instr = [
                f"Given the final code {mode_inputs}, provide the final {mode_outputs}.",
            ]

        return "\n".join(instr)


    def parse_code(self, code_data):
        code = (
            code_data.get("generated_code", "").split("---", 1)[0].split("\n\n\n", 1)[0]
        )
        code_match = re.search(r"```python[ \n](.*?)[ \n]```?", code, re.DOTALL)
        code_block = (code_match.group(1) if code_match else code).replace("\\n", "\n")
        if not code_block:
            return code, "Error: Empty code after parsing."
        if "\n" not in code_block and code_block.count("=") > 1:
            return code, "Error: Code format is not correct."
        lines = code_block.split("\n")
        last_line_match = re.match(r"^(\w+)\s*=", lines[-1].strip())
        if last_line_match and len(lines) > 1:
            code_block += "\n" + last_line_match.group(1)
        else:
            code_block = re.sub(
                r"([a-zA-Z_]\w* *=.*?)(?=[a-zA-Z_]\w* *=)", r"\1\n", code_block,
            )
            code_block = re.sub(
                r"([a-zA-Z_]\w* *=.*?)([a-zA-Z_]\w*)$", r"\1\n\2", code_block,
            )
        return code_block, None

    def execute_code(self, code):
        if not code:
            return code, None, "Error: Empty code before execution."
        interpreter = PythonInterpreter()
        try:
            output = str(interpreter.execute(code))
            return code, output, None
        except Exception as e:
            return code, None, str(e)
    def forward(self, **kwargs):
        input_kwargs = {
            field_name: kwargs[field_name] for field_name in self.input_fields
        }
        code_data = self.code_generate(**input_kwargs)
        parsed_code, error = self.parse_code(code_data)
        # FIXME: Don't try to execute the code if it didn't parse
        code, output, error = self.execute_code(parsed_code)
        hop = 0
        while hop < self.max_iters and error:
            print("Error in code execution")
            input_kwargs.update({"previous_code": code, "error": error})
            code_data = self.code_regenerate(**input_kwargs)
            parsed_code, error = self.parse_code(code_data)
            # FIXME: Don't try to execute the code if it didn't parse
            code, output, error = self.execute_code(parsed_code)
            hop += 1
            if hop == self.max_iters:
                print("Max hops reached. Error persists.")
                return None
        input_kwargs.update({"final_generated_code": code, "code_output": output})
        answer_gen_result = self.generate_answer(**input_kwargs)
        return answer_gen_result


import os
import queue
import random

def env_worker(inq, outq):
    """
    Worker process: creates a single AlfredTWEnv instance,
    handles 'init' (with task idx) and 'step' (with action).
    """

    try:
        import io
        import yaml
        import alfworld.agents.environment as environment
        from contextlib import redirect_stdout, redirect_stderr
    except ImportError:
        raise ImportError("alfworld is not installed. " \
            "Please install it via `pip install alfworld==0.3.5` then run `alfworld-download`.")

    buf = io.StringIO()
    base_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(base_dir, 'base_config.yml')

    with open(config_path) as f:
        config = yaml.safe_load(f)

    with redirect_stdout(buf), redirect_stderr(buf):
        base_env = environment.AlfredTWEnv(config, train_eval="train")

    env = None
    while True:
        cmd, data = inq.get()
        if cmd == 'init':
            env = base_env.init_env(batch_size=1)
            env.skip(data)
            task_def, info = env.reset()
            outq.put((task_def[0], info))
        elif cmd == 'step':
            obs, rew, done, info = env.step([data])
            outq.put((obs, rew, done, info))
        elif cmd == 'close':
            outq.put('CLOSED')
            break
        else:
            outq.put('UNKNOWN_CMD')


class EnvPool:
    """
    Pool of processes, each with a unique env_worker.
    Acquire a worker using a context manager for safe usage:
        with pool.session() as sess:
            sess.init(5)              # init with idx=5
            obs, rew, done, info = sess.step("go north")
            ...
    """
    def __init__(self, size=2):
        self.size = size
        self.workers = []
        self.available = queue.Queue()

        try:
            import multiprocess as mp
        except ImportError:
            raise ImportError("multiprocess is not installed. " \
                "Please install it via `pip install multiprocess`.")

        # Must call set_start_method('spawn') here, before creating any processes
        try:
            mp.set_start_method("spawn", force=True)
        except RuntimeError:
            # If it's already set, ignore
            pass

        ctx = mp.get_context("spawn")
        for i in range(size):
            inq = ctx.Queue()
            outq = ctx.Queue()
            p = ctx.Process(target=env_worker, args=(inq, outq), daemon=True)
            p.start()
            self.workers.append((inq, outq, p))
            self.available.put(i)

    def _acquire(self):
        wid = self.available.get()
        return wid, self.workers[wid][0], self.workers[wid][1]

    def _release(self, wid):
        self.available.put(wid)

    def close_all(self):
        """Close all processes in the pool."""
        while not self.available.empty():
            wid = self.available.get()
            inq, outq, proc = self.workers[wid]
            inq.put(('close', None))
            outq.get()  # Wait 'CLOSED'
            inq.close()
            outq.close()
            proc.join()

    def session(self):
        """Context manager that acquires/releases a single worker."""
        return _EnvSession(self)


class _EnvSession:
    """
    A context manager that acquires a worker from the pool,
    provides .init(idx) and .step(action), then releases the worker.
    """
    def __init__(self, pool: EnvPool):
        self.pool = pool
        self.wid = None
        self.inq = None
        self.outq = None

    def __enter__(self):
        self.wid, self.inq, self.outq = self.pool._acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pool._release(self.wid)

    def init(self, idx):
        self.inq.put(('init', idx))
        return self.outq.get()  # (task_def, info)

    def step(self, action):
        self.inq.put(('step', action))
        return self.outq.get()  # (obs, rew, done, info)


class AlfWorld:
    def __init__(self, max_threads=20):
        self.POOL = EnvPool(size=max_threads)

        import dspy
        dataset = [dspy.Example(idx=idx).with_inputs('idx') for idx in range(3500)]
        random.Random(0).shuffle(dataset)

        trainset, devset = dataset[:3000], dataset[-500:]
        assert len(trainset) + len(devset) <= len(dataset)

        self.trainset = trainset
        self.devset = devset

    def __del__(self):
        self.POOL.close_all()



import time
import socket
import requests
import threading
import subprocess

from datasets import Dataset
from typing import Any, Dict, List, Optional
from dspy.clients.provider import TrainingJob, Provider
from dspy.clients.utils_finetune import DataFormat, TrainingStatus, save_data


class LocalTrainingJob(TrainingJob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_path = None

    def cancel(self):
        raise NotImplementedError

    def status(self) -> TrainingStatus:
        raise NotImplementedError


class LocalProvider(Provider):
    def __init__(self):
        super().__init__()
        self.finetunable = True
        self.TrainingJob = LocalTrainingJob

    @staticmethod
    def launch(lm: "LM", launch_kwargs: Optional[Dict[str, Any]] = None):
        try:
            import sglang
        except ImportError:
            raise ImportError(
                "For local model launching, please install sglang by running "
                '`pip install "sglang[all]"; pip install flashinfer -i https://flashinfer.ai/whl/cu121/torch2.4/`'
            )

        if launch_kwargs is None:
            launch_kwargs = {}

        import os

        model = lm.model
        if model.startswith("openai/"):
            model = model[7:]
        if model.startswith("local:"):
            model = model[6:]

        print(f"Grabbing a free port to launch an SGLang server for model {model}")
        print(
            f"We see that CUDA_VISIBLE_DEVICES is {os.environ.get('CUDA_VISIBLE_DEVICES', 'unset')}"
        )
        port = get_free_port()
        timeout = launch_kwargs.get("timeout", 1800)
        command = f"python -m sglang.launch_server --model-path {model} --port {port} --host 0.0.0.0"

        # We will manually stream & capture logs.
        process = subprocess.Popen(
            command.replace("\\\n", " ").replace("\\", " ").split(),
            text=True,
            stdout=subprocess.PIPE,  # We'll read from pipe
            stderr=subprocess.STDOUT,  # Merge stderr into stdout
        )

        # A threading.Event to control printing after the server is ready.
        # This will store *all* lines (both before and after readiness).
        print(f"SGLang server process started with PID {process.pid}.")
        stop_printing_event = threading.Event()
        logs_buffer = []

        def _tail_process(proc, buffer, stop_event):
            while True:
                line = proc.stdout.readline()
                if not line and proc.poll() is not None:
                    # Process ended and no new line
                    break
                if line:
                    buffer.append(line)
                    # Print only if stop_event is not set
                    if not stop_event.is_set():
                        print(line, end="")

        # Start a background thread to read from the process continuously
        thread = threading.Thread(
            target=_tail_process,
            args=(process, logs_buffer, stop_printing_event),
            daemon=True,
        )
        thread.start()

        # Wait until the server is ready (or times out)
        base_url = f"http://localhost:{port}"
        try:
            wait_for_server(base_url, timeout=timeout)
        except TimeoutError:
            # If the server doesn't come up, we might want to kill it:
            process.kill()
            raise

        # Once server is ready, we tell the thread to stop printing further lines.
        stop_printing_event.set()

        # A convenience getter so the caller can see all logs so far (and future).
        def get_logs() -> str:
            # Join them all into a single string, or you might return a list
            return "".join(logs_buffer)

        # Let the user know server is up
        print(
            f"Server ready on random port {port}! Logs are available via lm.get_logs() method on returned lm."
        )

        lm.kwargs["api_base"] = f"http://localhost:{port}/v1"
        lm.kwargs["api_key"] = "local"
        lm.process = process
        lm.get_logs = get_logs

    @staticmethod
    def kill(model, kill_kwargs: Optional[Dict[str, Any]] = None):
        from sglang.utils import terminate_process

        terminate_process(model.process)

    @staticmethod
    def finetune(
        job: LocalTrainingJob,
        model: str,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[DataFormat] = None,
    ) -> str:
        data_path = save_data(train_data)

        if model.startswith("openai/"):
            model = model[7:]
        if model.startswith("local:"):
            model = model[6:]

        model_path = data_path.replace(".jsonl", f"__{model.replace('/', '__')}")
        print(f"[Local Provider] Data saved to {data_path}")

        default_train_kwargs = {
            "device": None,
            "use_peft": False,
            "num_train_epochs": 5,
            "per_device_train_batch_size": 1,
            "gradient_accumulation_steps": 8,
            "learning_rate": 1e-5,
            "max_seq_length": None,
            "packing": True,
            "bf16": True,
            "output_dir": model_path,
        }

        print(f"[Local Provider] Starting local training, will save to {model_path}")
        train_sft_locally(
            model_name=model,
            train_data=train_data,
            train_kwargs={**default_train_kwargs, **(train_kwargs or {})},
        )

        print("[Local Provider] Training complete")
        return f"openai/local:{model_path}"


def train_sft_locally(model_name, train_data, train_kwargs):
    try:
        import torch
        from transformers import AutoModelForCausalLM, AutoTokenizer
        from trl import SFTConfig, SFTTrainer, setup_chat_format
        from trl import apply_chat_template
    except ImportError:
        raise ImportError(
            "For local finetuning, please install torch, transformers, and trl "
            "by running `pip install -U torch transformers accelerate trl peft`"
        )

    device = train_kwargs.get("device", None)
    if device is None:
        device = (
            "cuda"
            if torch.cuda.is_available()
            else "mps" if torch.backends.mps.is_available() else "cpu"
        )
    print(f"Using device: {device}")

    model = AutoModelForCausalLM.from_pretrained(
        pretrained_model_name_or_path=model_name
    ).to(device)
    tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path=model_name)

    # Set up the chat format; generally only for non-chat model variants, hence the try-except.
    try:
        model, tokenizer = setup_chat_format(model=model, tokenizer=tokenizer)
    except Exception:
        pass

    if tokenizer.pad_token_id is None:
        print("Adding pad token to tokenizer")
        tokenizer.add_special_tokens({"pad_token": "[!#PAD#!]"})

    print("Creating dataset")
    trainset_dict = {
        "prompt": [entry["messages"][:-1] for entry in train_data],
        "completion": [[entry["messages"][-1]] for entry in train_data],
    }

    trainset = Dataset.from_dict(trainset_dict)
    trainset = trainset.map(apply_chat_template, fn_kwargs={"tokenizer": tokenizer})

    USE_PEFT = train_kwargs.get("use_peft", False)
    peft_config = None

    if USE_PEFT:
        from peft import LoraConfig

        rank_dimension = 32
        lora_alpha = 64
        lora_dropout = 0.05

        peft_config = LoraConfig(
            r=rank_dimension,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            bias="none",
            target_modules="all-linear",
            task_type="CAUSAL_LM",
        )

    sft_config = SFTConfig(
        output_dir=output_dir,
        num_train_epochs=train_kwargs["num_train_epochs"],
        per_device_train_batch_size=train_kwargs["per_device_train_batch_size"],
        gradient_accumulation_steps=train_kwargs["gradient_accumulation_steps"],
        learning_rate=train_kwargs["learning_rate"],
        max_grad_norm=2.0,
        logging_steps=20,
        warmup_ratio=0.03,
        lr_scheduler_type="constant",
        save_steps=10_000,
        bf16=train_kwargs["bf16"],
        max_seq_length=train_kwargs["max_seq_length"],
        packing=train_kwargs["packing"],
        dataset_kwargs={
            "add_special_tokens": False,  # Special tokens handled by template
            "append_concat_token": False,  # No additional separator needed
        },
    )

    print("Starting training")

    trainer = SFTTrainer(
        model=model,
        args=sft_config,
        train_dataset=trainset,
        peft_config=peft_config,
        processing_class=tokenizer,
    )

    # Train!
    trainer.train()

    # Save the model!
    trainer.save_model()

    MERGE = True
    if USE_PEFT and MERGE:
        from peft import AutoPeftModelForCausalLM

        # Load PEFT model on CPU
        model_ = AutoPeftModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=sft_config.output_dir,
            torch_dtype=torch.float16,
            low_cpu_mem_usage=True,
        )

        merged_model = model_.merge_and_unload()
        merged_model.save_pretrained(
            sft_config.output_dir, safe_serialization=True, max_shard_size="5GB"
        )

    # Clean up!
    import gc

    del model
    del tokenizer
    del trainer
    gc.collect()
    torch.cuda.empty_cache()

    return output_dir


def get_free_port() -> int:
    """
    Return a free TCP port on localhost.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]


def wait_for_server(base_url: str, timeout: int = None) -> None:
    """
    Wait for the server to be ready by polling the /v1/models endpoint.

    Args:
        base_url: The base URL of the server (e.g. http://localhost:1234)
        timeout: Maximum time to wait in seconds. None means wait forever.
    """
    start_time = time.time()
    while True:
        try:
            response = requests.get(
                f"{base_url}/v1/models",
                headers={"Authorization": "Bearer None"},
            )
            if response.status_code == 200:
                # A small extra sleep to ensure server is fully up.
                time.sleep(5)
                break

            if timeout and (time.time() - start_time) > timeout:
                raise TimeoutError("Server did not become ready within timeout period")
        except requests.exceptions.RequestException:
            # Server not up yet, wait and retry
            time.sleep(1)



import os
import warnings
from typing import Any, Literal, Optional, Union

import requests

import dspy
from dspy.dsp.utils import dotdict


class YouRM(dspy.Retrieve):
    """Retriever for You.com's Search and News API.

    [API reference](https://documentation.you.com/api-reference/)

    Args:
        ydc_api_key: you.com API key, if `YDC_API_KEY` is not set in the environment
        k: If ``endpoint="search"``, the max snippets to return per search hit.
           If ``endpoint="news"``, the max articles to return.
        endpoint: you.com endpoints
        num_web_results: The max number of web results to return, must be under 20
        safesearch: Safesearch settings, one of "off", "moderate", "strict", defaults to moderate
        country: Country code, ex: 'US' for United States, see API reference for more info
        search_lang: (News API) Language codes, ex: 'en' for English, see API reference for more info
        ui_lang: (News API) User interface language for the response, ex: 'en' for English.
                            See API reference for more info
        spellcheck: (News API) Whether to spell check query or not, defaults to True
    """

    def __init__(
        self,
        ydc_api_key: Optional[str] = None,
        k: int = 3,
        endpoint: Literal["search", "news"] = "search",
        num_web_results: Optional[int] = None,
        safesearch: Optional[Literal["off", "moderate", "strict"]] = None,
        country: Optional[str] = None,
        search_lang: Optional[str] = None,
        ui_lang: Optional[str] = None,
        spellcheck: Optional[bool] = None,
    ):
        super().__init__(k=k)

        # Data validation
        if not ydc_api_key and not os.environ.get("YDC_API_KEY"):
            raise RuntimeError('You must supply `ydc_api_key` or set environment variable "YDC_API_KEY"')

        if endpoint not in ("search", "news"):
            raise ValueError('`endpoint` must be either "search" or "news"')

        # Raise warning if News API-specific fields are set but endpoint is not "news"
        if endpoint != "news":
            news_api_fields = (search_lang, ui_lang, spellcheck)
            for field in news_api_fields:
                if field:
                    warnings.warn(
                        (
                            f"News API-specific field '{field}' is set but `{endpoint=}`. "
                            "This will have no effect."
                        ),
                        UserWarning,
                    )

        self.ydc_api_key = ydc_api_key or os.environ.get("YDC_API_KEY")
        self.endpoint = endpoint
        self.num_web_results = num_web_results
        self.safesearch = safesearch
        self.country = country
        self.search_lang = search_lang
        self.ui_lang = ui_lang
        self.spellcheck = spellcheck

    def _generate_params(self, query: str) -> dict[str, Any]:
        params = {"safesearch": self.safesearch, "country": self.country}

        if self.endpoint == "search":
            params.update(
                query=query,
                num_web_results=self.num_web_results,
            )
        elif self.endpoint == "news":
            params.update(
                q=query,
                count=self.num_web_results,
                search_lang=self.search_lang,
                ui_lang=self.ui_lang,
                spellcheck=self.spellcheck,
            )

        # Remove `None` values
        params = {k: v for k, v in params.items() if v is not None}
        return params

    def forward(self, query_or_queries: Union[str, list[str]], k: Optional[int] = None) -> dspy.Prediction:
        k = k if k is not None else self.k

        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        docs: list[str]
        for query in queries:
            headers = {"X-API-Key": self.ydc_api_key}
            params = self._generate_params(query)
            response = requests.get(
                f"https://api.ydc-index.io/{self.endpoint}",
                params=params,
                headers=headers,
            )
            response.raise_for_status()
            results = response.json()

            if self.endpoint == "search":
                docs = [snippet for hits in results["hits"][:k] for snippet in hits["snippets"]]
            elif self.endpoint == "news":
                docs = [article["description"] for article in results["news"]["results"][:k]]
        return [dotdict({"long_text": document}) for document in docs]



import os
from typing import List, Optional, Union
import string
import random

import backoff
from openai import (
    APITimeoutError,
    InternalServerError,
    OpenAI,
    RateLimitError,
    UnprocessableEntityError,
)

from dspy import Retrieve, Prediction
from dspy.dsp.utils.settings import settings
from dspy.dsp.utils import dotdict

try:
    import falkordb
except ImportError:
    raise ImportError(
        "Please install the falkordb package by running `pip install dspy-ai[falkordb]`"
    )
import redis.exceptions


def generate_random_string(length: int) -> str:
    characters = string.ascii_letters
    random_string = "".join(random.choice(characters) for _ in range(length))
    return random_string


class Embedder:
    def __init__(self, provider: str, model: str):
        self.provider = provider
        if self.provider == "openai":
            api_key = os.getenv("OPENAI_API_KEY")
            if not api_key:
                raise ValueError(
                    "Environment variable OPENAI_API_KEY must be set to"
                    "use openai as embedding provider"
                )
            self.client = OpenAI()
            self.model = model
        else:
            raise ValueError(f"Unsupported provider: {provider}")

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
        ),
        max_time=settings.backoff_time,
    )
    def __call__(self, queries: Union[str, List[str]]) -> List[List[float]]:
        if isinstance(queries, str):
            queries = [queries]

        if self.provider == "openai":
            embedding = self.client.embeddings.create(input=queries, model=self.model)
            return [result.embedding for result in embedding.data]


DEFAULT_INDEX_QUERY = "CALL db.idx.vector.queryNodes($node_label, $embedding_node_property, $k, vecf32($embedding)) YIELD node, score "


class FalkordbRM(Retrieve):
    """
    Implements a retriever that utilizes FalkorDB for retrieving passages.
    This class manages a connection to a FalkorDB database using official FalkorDB Python drivers and requires
    the database credentials. That is, if using a local FalkorDB session, host and port else if using a FalkorDB cloud session,
    host, port, username, and password to be set as environment variables and optionally the database name.
    Additionally, it utilizes an embedding provider (defaulting to OpenAI's services) to compute query embeddings,
    which are then used to find the most relevant nodes in the FalkorDB graph based on the specified node property or custom retrieval query.

    Returns a list of passages in the form of `dspy.Prediction` objects

    Args:
        Args:
        node_label (str): The label of the node in the FalkorDB database to query against
        text_node_property (str): The property of the node containing the text.
        embedding_node_property (List[float]): The property of the node containing the embeddings.
        k (Optional[int]): The default number of top passages to retrieve. Defaults to 5.
        retrieval_query (Optional[str]): Custom Cypher query for retrieving passages.
        embedding_provider (str): The provider of the embedding service. Defaults to "openai".
        embedding_model (str): The model identifier for generating embeddings. Defaults to "text-embedding-ada-002".

    Examples:
        Below is a code snippet showcasing how to initialize FalkordbRM with environment variables for the database connection and OpenAI as the embedding provider:

        ```python
        import os

        import dspy
        import openai

        os.environ["FALKORDB_HOST"] = "localhost"
        os.environ["FALORDB_PORT"] = "6379"
        os.environ["OPENAI_API_KEY"] = "sk-" (Only if using openai as embedding's provider)

        # Uncomment and set the following if you are using FalkorDB cloud
        # os.environ["FALKORDB_USERNAME"] = "falkordb"
        # os.environ["FALKORDB_PASSWORD"] = "password"


        falkordb_retriever = FalkordbRM(
            node_label="myIndex",
            text_node_property="text",
            k=10,
            embedding_provider="openai",
            embedding_model="text-embedding-ada-002",
        )

        dspy.settings.configure(rm=falkordb_retriever)
        ```

        In this example, `FalkordbRM` is configured to retrieve nodes based on the "text" property from an index on a node labeled "myIndex",
        using embeddings computed by OpenAI's "text-embedding-ada-002" model.
    """

    def __init__(
        self,
        node_label: str,
        text_node_property: str = None,
        embedding_node_property: str = None,
        k: int = 5,
        retrieval_query: Optional[str] = None,
        embedding_provider: str = "openai",
        embedding_model: str = "text-embedding-ada-002",
    ):
        super().__init__(k=k)
        self.node_label = node_label
        self.username = os.getenv("FALKORDB_USERNAME", None)
        self.password = os.getenv("FALKORDB_PASSWORD", None)
        self.host = os.getenv("FALKORDB_HOST", "localhost")
        self.port = int(os.getenv("FALKORDB_PORT", 6379))

        self.database = os.getenv("FALKORDB_DATABASE", generate_random_string(4))
        self.k = k
        self.retrieval_query = retrieval_query
        self.text_node_property = text_node_property
        self.embedding_node_property = embedding_node_property
        if not self.text_node_property and not self.retrieval_query:
            raise ValueError(
                "Either `text_node_property` or `retrieval_query` must be set"
            )
        if not embedding_node_property:
            raise ValueError("`embedding_node_property` must be set")
        try:
            self.driver = falkordb.FalkorDB(
                host=self.host,
                port=self.port,
                username=self.username,
                password=self.password,
            ).select_graph(self.database)

        except (
            redis.exceptions.ConnectionError,
            redis.exceptions.AuthenticationError,
        ) as e:
            raise ConnectionError("Failed to connect to FalkorDB database") from e

        self.embedder = Embedder(provider=embedding_provider, model=embedding_model)

    def forward(
        self, query_or_queries: Union[str, List[str]], k: Optional[int]
    ) -> Prediction:
        if not isinstance(query_or_queries, list):
            query_or_queries = [query_or_queries]
        query_vectors = self.embedder(query_or_queries)
        contents = []
        retrieval_query = (
            self.retrieval_query
            or f"RETURN node.{self.text_node_property} AS text, score"
        )
        if not k:
            k = self.k

        for vector in query_vectors:
            params = {
                "embedding": vector,
                "node_label": self.node_label,
                "text_node_property": self.text_node_property,
                "embedding_node_property": self.embedding_node_property,
                "k": k,
            }
            try:
                records = self.driver.query(
                    DEFAULT_INDEX_QUERY + retrieval_query,
                    params=params,
                ).result_set
            except Exception as e:
                if "Invalid arguments" in str(e):
                    raise ValueError(
                        f"There is no vector index on node label, {self.node_label}"
                        f" and node property, {self.embedding_node_property}"
                    )
            contents.extend(
                [
                    {"passage": dotdict({"long_text": r[1]}), "score": r[0]}
                    for r in records
                ]
            )
        sorted_passages = sorted(
            contents,
            key=lambda x: x["score"],
            reverse=True,
        )[:k]
        return [el["passage"] for el in sorted_passages]



from asyncio import iscoroutinefunction
from typing import Any, AsyncGenerator, Awaitable, Callable

import litellm
import ujson
from anyio import create_memory_object_stream, create_task_group
from anyio.streams.memory import MemoryObjectSendStream

from dspy.primitives.prediction import Prediction
from dspy.primitives.program import Module
from dspy.utils.asyncify import asyncify


def streamify(program: Module) -> Callable[[Any, Any], Awaitable[Any]]:
    """
    Wrap a DSPy program so that it streams its outputs incrementally, rather than returning them
    all at once.

    Args:
        program: The DSPy program to wrap with streaming functionality.

    Returns:
        A function that takes the same arguments as the original program, but returns an async
            generator that yields the program's outputs incrementally.

    Example:

    ```python
    class TestSignature(dspy.Signature):
        input_text: str = dspy.InputField()
        output_text: str = dspy.OutputField()

    # Create the program and wrap it with streaming functionality
    program = dspy.streamify(dspy.Predict(TestSignature))

    # Use the program with streaming output
    async def use_streaming():
        output_stream = program(input_text="Test")
        async for value in output_stream:
            print(value)  # Print each streamed value incrementally
    ```
    """
    import dspy

    if not iscoroutinefunction(program):
        program = asyncify(program)

    async def generator(args, kwargs, stream: MemoryObjectSendStream):
        with dspy.settings.context(send_stream=stream):
            prediction = await program(*args, **kwargs)

        await stream.send(prediction)

    async def streamer(*args, **kwargs):
        send_stream, receive_stream = create_memory_object_stream(16)
        async with create_task_group() as tg, send_stream, receive_stream:
            tg.start_soon(generator, args, kwargs, send_stream)

            async for value in receive_stream:
                yield value
                if isinstance(value, Prediction):
                    return

    return streamer


async def streaming_response(streamer: AsyncGenerator) -> AsyncGenerator:
    """
    Convert a DSPy program output stream to an OpenAI-compatible output stream that can be
    used by a service as an API response to a streaming request.

    Args:
        streamer: An async generator that yields values from a DSPy program output stream.
    Returns:
        An async generator that yields OpenAI-compatible streaming response chunks.
    """
    async for value in streamer:
        if isinstance(value, Prediction):
            data = {"prediction": {k: v for k, v in value.items(include_dspy=False)}}
            yield f"data: {ujson.dumps(data)}\n\n"
        elif isinstance(value, litellm.ModelResponseStream):
            data = {"chunk": value.json()}
            yield f"data: {ujson.dumps(data)}\n\n"
        elif isinstance(value, str) and value.startswith("data:"):
            # The chunk value is an OpenAI-compatible streaming chunk value,
            # e.g. "data: {"finish_reason": "stop", "index": 0, "is_finished": True, ...}",
            # so yield it directly
            yield value
        else:
            raise ValueError(f"Unknown chunk value type: {value}")
    yield "data: [DONE]\n\n"



import dspy
from dspy.primitives.program import Module
from dspy.signatures.signature import ensure_signature


class ChainOfThought(Module):
    def __init__(self, signature, rationale_type=None, **config):
        super().__init__()

        signature = ensure_signature(signature)

        prefix = "Reasoning: Let's think step by step in order to"
        desc = "${reasoning}"
        rationale_type = rationale_type or dspy.OutputField(prefix=prefix, desc=desc)
        extended_signature = signature.prepend("reasoning", rationale_type, type_=str)
        
        self.predict = dspy.Predict(extended_signature, **config)

    def forward(self, **kwargs):
        return self.predict(**kwargs)



from abc import ABC, abstractmethod

GLOBAL_HISTORY = []

class BaseLM(ABC):
    def __init__(self, model, model_type='chat', temperature=0.0, max_tokens=1000, cache=True, **kwargs):
        self.model = model
        self.model_type = model_type
        self.cache = cache
        self.kwargs = dict(temperature=temperature, max_tokens=max_tokens, **kwargs)
        self.history = []

    @abstractmethod
    def __call__(self, prompt=None, messages=None, **kwargs):
        pass

    def inspect_history(self, n: int = 1):
        _inspect_history(self.history, n)

    def update_global_history(self, entry):
        GLOBAL_HISTORY.append(entry)


def _green(text: str, end: str = "\n"):
    return "\x1b[32m" + str(text).lstrip() + "\x1b[0m" + end


def _red(text: str, end: str = "\n"):
    return "\x1b[31m" + str(text) + "\x1b[0m" + end

def _blue(text: str, end: str = "\n"):
    return "\x1b[34m" + str(text) + "\x1b[0m" + end


def _inspect_history(history, n: int = 1):
    """Prints the last n prompts and their completions."""

    for item in history[-n:]:
        messages = item["messages"] or [{"role": "user", "content": item["prompt"]}]
        outputs = item["outputs"]
        timestamp = item.get("timestamp", "Unknown time")

        print("\n\n\n")
        print("\x1b[34m" + f"[{timestamp}]" + "\x1b[0m" + "\n")

        for msg in messages:
            print(_red(f"{msg['role'].capitalize()} message:"))
            if isinstance(msg["content"], str):
                print(msg["content"].strip())
            else:
                if isinstance(msg["content"], list):
                    for c in msg["content"]:
                        if c["type"] == "text":
                            print(c["text"].strip())
                        elif c["type"] == "image_url":
                            image_str = ""
                            if "base64" in c["image_url"].get("url", ""):
                                len_base64 = len(c["image_url"]["url"].split("base64,")[1])
                                image_str = f"<{c['image_url']['url'].split('base64,')[0]}base64,<IMAGE BASE 64 ENCODED({str(len_base64)})>"
                            else:
                                image_str = f"<image_url: {c['image_url']['url']}>"
                            print(_blue(image_str.strip()))
            print("\n")

        print(_red("Response:"))
        print(_green(outputs[0].strip()))

        if len(outputs) > 1:
            choices_text = f" \t (and {len(outputs)-1} other completions)"
            print(_red(choices_text, end=""))

    print("\n\n\n")

def inspect_history(n: int = 1):
    """The global history shared across all LMs."""
    return _inspect_history(GLOBAL_HISTORY, n)


from collections import defaultdict
from typing import List, Union

import dspy
from dspy.dsp.utils import dotdict

try:
    import marqo
except ImportError:
    raise ImportError(
        "The 'marqo' extra is required to use MarqoRM. Install it with `pip install dspy-ai[marqo]`",
    )

class MarqoRM(dspy.Retrieve):
    """
    A retrieval module that uses Marqo to return the top passages for a given query.

    Assumes that a Marqo index has been created and populated with the following payload:
        - document: The text of the passage

    Args:
        marqo_index_name (str): The name of the marqo index.
        marqo_client (marqo.client.Client): A marqo client instance.
        k (int, optional): The number of top passages to retrieve. Defaults to 3.
        page_content (str, optional): The name of the field in the marqo index that contains the text of the passage. Defaults to 'document'.
        filter_string (str, optional): A filter string to use when searching. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the marqo search function.

    Examples:
        Below is a code snippet that shows how to use Marqo as the default retriver:
        ```python
        import marqo
        marqo_client = marqo.Client(url="http://0.0.0.0:8882")

        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        retriever_model = MarqoRM("my_index_name", marqo_client=marqo_client)
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use Marqo in the forward() function of a module
        ```python
        self.retrieve = MarqoRM("my_index_name", marqo_client=marqo_client, k=num_passages)
        ```
    """

    def __init__(
        self,
        marqo_index_name: str,
        marqo_client: marqo.client.Client,
        k: int = 3,
        page_content: str = 'document',
        filter_string: str = None,
    ):
        self._marqo_index_name = marqo_index_name
        self._marqo_client = marqo_client
        self.page_content = page_content
        self.filter_string = filter_string

        super().__init__(k=k)

    def forward(self, query_or_queries: Union[str, List[str]], k=None, **kwargs) -> dspy.Prediction:
        """Search with Marqo for self.k top passages for query

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [q for q in queries if q]
        limit = k if k else self.k

        all_query_results = []
        for query in queries:
            _result = self._marqo_client.index(self._marqo_index_name).search(
                q=query,
                limit=limit,
                filter_string=self.filter_string,
                **kwargs,
            )
            all_query_results.append(_result)

        passages = defaultdict(float)

        for result_dict in all_query_results:
            for result in result_dict['hits']:
                passages[result[self.page_content]] += result['_score']

        sorted_passages = sorted(
            passages.items(), key=lambda x: x[1], reverse=True)[:limit]
        return [dotdict({"long_text": passage}) for passage, _ in sorted_passages]



import sys
import tqdm
import signal
import logging
import threading
import traceback
import contextlib
from tqdm.contrib.logging import logging_redirect_tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed

logger = logging.getLogger(__name__)


class ParallelExecutor:
    def __init__(
        self,
        num_threads,
        max_errors=5,
        disable_progress_bar=False,
        provide_traceback=False,
        compare_results=False,
    ):
        """Offers isolation between the tasks (dspy.settings) irrespective of whether num_threads == 1 or > 1."""
        self.num_threads = num_threads
        self.disable_progress_bar = disable_progress_bar
        self.max_errors = max_errors
        self.provide_traceback = provide_traceback
        self.compare_results = compare_results

        self.error_count = 0
        self.error_lock = threading.Lock()
        self.cancel_jobs = threading.Event()

    def execute(self, function, data):
        wrapped_function = self._wrap_function(function)
        if self.num_threads == 1:
            return self._execute_isolated_single_thread(wrapped_function, data)
        else:
            return self._execute_multi_thread(wrapped_function, data)

    def _wrap_function(self, function):
        # Wrap the function with error handling
        def wrapped(item):
            if self.cancel_jobs.is_set():
                return None
            try:
                return function(item)
            except Exception as e:
                with self.error_lock:
                    self.error_count += 1
                    current_error_count = self.error_count
                if current_error_count >= self.max_errors:
                    self.cancel_jobs.set()
                    raise e
                if self.provide_traceback:
                    logger.error(
                        f"Error processing item {item}: {e}\nStack trace:\n{traceback.format_exc()}"
                    )
                else:
                    logger.error(
                        f"Error processing item {item}: {e}. Set `provide_traceback=True` to see the stack trace."
                    )
                return None
        return wrapped

    def _execute_isolated_single_thread(self, function, data):
        results = []
        pbar = tqdm.tqdm(
            total=len(data),
            dynamic_ncols=True,
            disable=self.disable_progress_bar,
            file=sys.stdout
        )

        from dspy.dsp.utils.settings import thread_local_overrides
        original_overrides = thread_local_overrides.overrides

        for item in data:
            with logging_redirect_tqdm():
                if self.cancel_jobs.is_set():
                    break

                # Create an isolated context for each task by copying current overrides
                # This way, even if an iteration modifies the overrides, it won't affect subsequent iterations
                thread_local_overrides.overrides = original_overrides.copy()

                try:
                    result = function(item)
                    results.append(result)
                finally:
                    thread_local_overrides.overrides = original_overrides

                if self.compare_results:
                    # Assumes score is the last element of the result tuple
                    self._update_progress(
                        pbar,
                        sum([r[-1] for r in results if r is not None]),
                        len([r for r in data if r is not None]),
                    )
                else:
                    self._update_progress(pbar, len(results), len(data))

        pbar.close()

        if self.cancel_jobs.is_set():
            logger.warning("Execution was cancelled due to errors.")
            raise Exception("Execution was cancelled due to errors.")

        return results

    def _update_progress(self, pbar, nresults, ntotal):
        if self.compare_results:
            percentage = round(100 * nresults / ntotal, 1) if ntotal > 0 else 0
            pbar.set_description(f"Average Metric: {nresults:.2f} / {ntotal} ({percentage}%)")
        else:
            pbar.set_description(f"Processed {nresults} / {ntotal} examples")

        pbar.update()

    def _execute_multi_thread(self, function, data):
        results = [None] * len(data)  # Pre-allocate results list to maintain order
        job_cancelled = "cancelled"

        @contextlib.contextmanager
        def interrupt_handler_manager():
            """Sets the cancel_jobs event when a SIGINT is received, only in the main thread."""

            # TODO: Is this check conducive to nested usage of ParallelExecutor?
            if threading.current_thread() is threading.main_thread():
                default_handler = signal.getsignal(signal.SIGINT)

                def interrupt_handler(sig, frame):
                    self.cancel_jobs.set()
                    logger.warning("Received SIGINT. Cancelling execution.")
                    # Re-raise the signal to allow default behavior
                    default_handler(sig, frame)

                signal.signal(signal.SIGINT, interrupt_handler)
                try:
                    yield
                finally:
                    signal.signal(signal.SIGINT, default_handler)
            else:
                # If not in the main thread, skip setting signal handlers
                yield

        def cancellable_function(parent_overrides, index_item):
            index, item = index_item
            if self.cancel_jobs.is_set():
                return index, job_cancelled

            # Create an isolated context for each task by copying parent's overrides
            from dspy.dsp.utils.settings import thread_local_overrides
            original_overrides = thread_local_overrides.overrides
            thread_local_overrides.overrides = parent_overrides.copy()

            try:
                return index, function(item)
            finally:
                thread_local_overrides.overrides = original_overrides

        with ThreadPoolExecutor(max_workers=self.num_threads) as executor, interrupt_handler_manager():
            from dspy.dsp.utils.settings import thread_local_overrides
            parent_overrides = thread_local_overrides.overrides.copy()

            futures = {}
            for pair in enumerate(data):
                # Pass the parent thread's overrides to each thread
                future = executor.submit(cancellable_function, parent_overrides, pair)
                futures[future] = pair

            pbar = tqdm.tqdm(
                total=len(data),
                dynamic_ncols=True,
                disable=self.disable_progress_bar,
                file=sys.stdout
            )

            for future in as_completed(futures):
                index, result = future.result()

                if result is job_cancelled:
                    continue

                results[index] = result

                if self.compare_results:
                    # Assumes score is the last element of the result tuple
                    self._update_progress(
                        pbar,
                        sum([r[-1] for r in results if r is not None]),
                        len([r for r in results if r is not None]),
                    )
                else:
                    self._update_progress(
                        pbar,
                        len([r for r in results if r is not None]),
                        len(data),
                    )

            pbar.close()

        if self.cancel_jobs.is_set():
            logger.warning("Execution was cancelled due to errors.")
            raise Exception("Execution was cancelled due to errors.")

        return results



"""
Retriever model for LanceDB
Author: Prashant Dixit (@PrashantDixit0)
"""

from typing import List, Union

import backoff
import lancedb.table

from dspy import Retrieve, Prediction
from dspy.dsp.utils.settings import settings
from dspy.dsp.utils import dotdict

try:
    import lancedb
except ImportError:
    lancedb = None

if lancedb is None:
    raise ImportError(
        "The lancedb library is required to use LancedbRM. Install it with `pip install dspy-ai[lancedb]`",
    )


import openai

try:
    OPENAI_LEGACY = int(openai.version.__version__[0]) == 0
except Exception:
    OPENAI_LEGACY = True

try:
    import openai.error
    ERRORS = (openai.error.RateLimitError, openai.error.ServiceUnavailableError, openai.error.APIError)
except Exception:
    ERRORS = (openai.RateLimitError, openai.APIError)

    
class LancedbRM(Retrieve):
    """
    A retrieval module that uses LanceDB to return the top passages for a given query.

    Assumes that the LanceDB table has been created and populated with the following metadata:
        - text: The text of the passage

    Args:
        table_name (str): The name of the table to query against.
        persist_directory (str): directory where database is stored.
        k (int, optional): The number of top passages to retrieve. Defaults to 3.

    Returns:
        dspy.Prediction: An object containing the retrieved passages.

    Examples:
        Below is a code snippet that shows how to use this as the default retriever:
        ```python
        llm = dspy.OpenAI(model="gpt-3.5-turbo")
        retriever_model = LancedbRM()
        dspy.settings.configure(lm=llm, rm=retriever_model)
        ```

        Below is a code snippet that shows how to use this in the forward() function of a module
        ```python
        self.retrieve = LancedbRM(k=num_passages)
        ```
    """

    def __init__(
        self,
        table_name: str,
        persist_directory: str,
        k: int = 3,
    ):
        
        self._table = self._init_lancedb(
            table_name, persist_directory,
        )

        super().__init__(k=k)

    def _init_lancedb(
        self,
        table_name: str,
        persist_directory: str,
    ):
        """
            Initialize LanceDB database and return the table instance.

            Args:
                table_name (str): The name of the table to load. If the table does not exist, it will be Raise Value Error.

            Raises:
                ValueError: If table is already not created

            Returns:
                Lancedb.table: The loaded table.
        """

        self.db = lancedb.connect(persist_directory)
        table_names = self.db.table_names()

        if table_name not in table_names:
            raise ValueError(
                f"Table does not exist in lancedb database persist_directory. Create table with name {table_name}, then reintialize retriever instance",
            )
        else:
            table = self.db.open_table(table_name)
        return table
    
    
    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
    )

    def forward(self, query_or_queries: Union[str, List[str]]) -> Prediction:
        """Search with Lancedb for self.k top passages for query

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = (
            [query_or_queries]
            if isinstance(query_or_queries, str)
            else query_or_queries
        )
        queries = [q for q in queries if q]  # Filter empty queries
        
        passages = []
        for q in queries:
            results = self._table.search(q).limit(self.k).to_list()
            for r in results:
                passages.append(r['text'])
        return Prediction(passages=[dotdict({"long_text": passage}) for passage in passages])



import dspy

try:
    import graphviz  # type: ignore
    graphviz_available = True
except ImportError:
    graphviz_available = False


class ModuleGraph:
    def __init__(self, module_name, module):
        if graphviz_available is False:
            raise ImportError(
                """Please install graphviz to use this feature.
                Run 'pip install graphviz'""")

        self.graph = graphviz.Digraph(format='png')
        self.nodes = set()
        self.module_name = module_name
        self.module = module
        self.inspect_settings(dspy.settings)
        self.add_module(self.module_name, self.module)

    def inspect_settings(self, settings):
        """Check for the existence and configuration of LM and RM and add them to the graph."""
        components = {'lm': settings.lm, 'rm': settings.rm}
        for component_name, component in components.items():
            if component:
                details = {attr: getattr(component, attr) for attr in dir(component)
                           if not attr.startswith('_') and not callable(getattr(component, attr))}
                component_details = f"{component_name.upper()} Details: " + ', '.join(f"{k}: {v}" for k, v in details.items() if k!='history')
                self.graph.node(component_name, label=component_details, shape='box')
                self.nodes.add(component_name)

    def add_module(self, module_name, module):
        """Add a module to the graph"""

        module_type = type(module)

        if 'dspy.predict' in str(module_type):
            module_name = self.generate_module_name(module_name, module_type)
            self.process_submodule(module_name, module)
        else:
            self.process_submodules(module_name, module)

    def generate_module_name(self, base_name, module_type):  # noqa: ANN201
        """ Generate a module name based on the module type"""
        type_map = {
            'Predict': '__Predict',
            'ReAct': '__ReAct',
            'ChainOfThought': '__ChainOfThought',
            'ProgramOfThought': '__ProgramOfThought',
            'MultiChainComparison': '__MultiChainComparison',
            'majority': '__majority',
        }

        for key, suffix in type_map.items():
            if key in str(module_type):
                return base_name + suffix
        return base_name

    def process_submodules(self, module_name, module):
        """ Process submodules of a module and add them to the graph"""

        for sub_module_name, sub_module in module.__dict__.items():
            if isinstance(sub_module, dspy.Predict):
                sub_module_name = self.generate_module_name(sub_module_name, type(sub_module))
                self.process_submodule(f"{module_name}__{sub_module_name}", sub_module)
            elif isinstance(sub_module, (dspy.Module, dspy.Retrieve)):
                self.add_module(f"{module_name}__{sub_module_name}", sub_module)
                if isinstance(sub_module, dspy.Retrieve):
                    self.graph.edge("rm", 'lm', label='RM used in Module')

    def process_submodule(self, sub_module_name, sub_module):
        """Process a submodule and add it to the graph"""

        for field_type, fields in [('input', sub_module.signature.input_fields),
                                   ('output', sub_module.signature.output_fields)]:
            for field_name, field in fields.items():
                node_id = f"{sub_module_name}_{field_type}_{field_name}"
                if node_id not in self.nodes:
                    label = f"{field_name}: ({field.json_schema_extra['desc']})"
                    self.graph.node(node_id, label=label, shape='ellipse')
                    self.nodes.add(node_id)
                edge_direction = (node_id, sub_module_name) if field_type == 'input' else (sub_module_name, node_id)
                self.graph.edge(*edge_direction)

        # Add node for the submodule itself
        self.graph.node(sub_module_name, label=sub_module_name, shape='box')
        self.nodes.add(sub_module_name)

        # Connect submodule to LM if configured
        if 'lm' in self.nodes:
            self.graph.edge('lm', sub_module_name, label='LM used in Module')

    def render_graph(self, filename=None):
        """Render the graph to a file(png)"""
        if filename is None:
            filename = self.module_name
        self.graph.render(filename)


# Example usage of the ModuleGraph class:
# import dspy
# import os
# from dotenv import load_dotenv
# from dspy.experimental import ModuleGraph

# load_dotenv()

# # Configuration of dspy models
# llm = dspy.OpenAI(
#     model='gpt-3.5-turbo',
#     api_key=os.environ['OPENAI_API_KEY'],
#     max_tokens=100
# )

# colbertv2_wiki = dspy.ColBERTv2(url='http://20.102.90.50:2017/wiki17_abstracts')

# dspy.settings.configure(lm=llm, rm=colbertv2_wiki)

# class GenerateAnswer(dspy.Signature):
#   "Answer with long and detailed answers"
#   context = dspy.InputField(desc="may content relevant facts")
#   question = dspy.InputField()
#   answer = dspy.OutputField(desc="often between 10 and 50 words")

# class RAG(dspy.Module):
#   def __init__(self, num_passages=3):
#     super().__init__()
#     self.retrieve = dspy.Retrieve(k=num_passages)
#     self.generate_answer = dspy.ChainOfThought(GenerateAnswer)

#   def forward(self, question):
#     context = self.retrieve(question).passages
#     prediction = self.generate_answer(context=context, question=question)
#     return dspy.Prediction(context=context, answer=prediction.answer)

# rag_system = RAG()
# graph = ModuleGraph("RAG", rag_system)

# graph.render_graph()



from typing import List, Optional, Union

import dspy
from dspy.dsp.utils import dotdict
from dspy.primitives.prediction import Prediction

try:
    import weaviate
    from weaviate.util import get_valid_uuid
    from uuid import uuid4
except ImportError as err:
    raise ImportError(
        "The 'weaviate' extra is required to use WeaviateRM. Install it with `pip install dspy-ai[weaviate]`",
    ) from err


class WeaviateRM(dspy.Retrieve):
    """A retrieval module that uses Weaviate to return the top passages for a given query.

    Assumes that a Weaviate collection has been created and populated with the following payload:
        - content: The text of the passage

    Args:
        weaviate_collection_name (str): The name of the Weaviate collection.
        weaviate_client (WeaviateClient): An instance of the Weaviate client.
        k (int, optional): The default number of top passages to retrieve. Default to 3.

    Examples:
        Below is a code snippet that shows how to use Weaviate as the default retriever:
        ```python
        import weaviate

        llm = dspy.Cohere(model="command-r-plus", api_key=api_key)
        weaviate_client = weaviate.connect_to_[local, wcs, custom, embedded]("your-path-here")
        retriever_model = WeaviateRM("my_collection_name", weaviate_client=weaviate_client)
        dspy.settings.configure(lm=llm, rm=retriever_model)

        retrieve = dspy.Retrieve(k=1)
        topK_passages = retrieve("what are the stages in planning, sanctioning and execution of public works").passages
        ```

        Below is a code snippet that shows how to use Weaviate in the forward() function of a module
        ```python
        self.retrieve = WeaviateRM("my_collection_name", weaviate_client=weaviate_client, k=num_passages)
        ```
    """

    def __init__(
        self,
        weaviate_collection_name: str,
        weaviate_client: Union[weaviate.WeaviateClient, weaviate.Client],
        weaviate_collection_text_key: Optional[str] = "content",
        k: int = 3,
    ):
        self._weaviate_collection_name = weaviate_collection_name
        self._weaviate_client = weaviate_client
        self._weaviate_collection = self._weaviate_client.collections.get(self._weaviate_collection_name)
        self._weaviate_collection_text_key = weaviate_collection_text_key

        # Check the type of weaviate_client (this is added to support v3 and v4)
        if hasattr(weaviate_client, "collections"):
            self._client_type = "WeaviateClient"
        elif hasattr(weaviate_client, "query"):
            self._client_type = "Client"
        else:
            raise ValueError("Unsupported Weaviate client type")

        super().__init__(k=k)

    def forward(self, query_or_queries: Union[str, List[str]], k: Optional[int] = None, **kwargs) -> Prediction:
        """Search with Weaviate for self.k top passages for query or queries.

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.
            kwargs :

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        k = k if k is not None else self.k
        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        queries = [q for q in queries if q]
        passages, parsed_results = [], []
        for query in queries:
            if self._client_type == "WeaviateClient":
                results = self._weaviate_collection.query.hybrid(
                    query=query,
                    limit=k,
                    **kwargs,
                )

                parsed_results = [result.properties[self._weaviate_collection_text_key] for result in results.objects]

            elif self._client_type == "Client":
                results = (
                    self._weaviate_client.query.get(
                        self._weaviate_collection_name,
                        [self._weaviate_collection_text_key],
                    )
                    .with_hybrid(query=query)
                    .with_limit(k)
                    .do()
                )

                results = results["data"]["Get"][self._weaviate_collection_name]
                parsed_results = [result[self._weaviate_collection_text_key] for result in results]

            passages.extend(dotdict({"long_text": d}) for d in parsed_results)

        return passages
    
    def get_objects(self, num_samples: int, fields: List[str]) -> List[dict]:
        """Get objects from Weaviate using the cursor API."""
        if self._client_type == "WeaviateClient":
            objects = []
            counter = 0
            for item in self._weaviate_collection.iterator():
                if counter >= num_samples:
                    break
                new_object = {}
                for key in item.properties.keys():
                    if key in fields:
                      new_object[key] = item.properties[key]
                objects.append(new_object)
                counter += 1
            return objects
        else:
            raise ValueError("`get_objects` is not supported for the v3 Weaviate Python client, please upgrade to v4.")
        
    def insert(self, new_object_properties: dict):
        if self._client_type == "WeaviateClient":
            self._weaviate_collection.data.insert(
                properties=new_object_properties,
                uuid=get_valid_uuid(uuid4())
            )
        else:
            raise AttributeError("`insert` is not supported for the v3 Weaviate Python client, please upgrade to v4.")



# from dspy.teleprompt.mipro_optimizer import MIPRO

# """
# ===============================================================
# DEPRECATED!!!
# PLEASE USE MIPRO INSTEAD.
# ===============================================================

# USAGE SUGGESTIONS:

# The following code can be used to compile a optimized signature teleprompter using the BayesianSignatureOptimizer, and evaluate it on an end task:

# from dspy.teleprompt import BayesianSignatureOptimizer

# teleprompter = BayesianSignatureOptimizer(prompt_model=prompt_model, task_model=task_model, metric=metric, n=10, init_temperature=1.0)
# kwargs = dict(num_threads=NUM_THREADS, display_progress=True, display_table=0)
# compiled_prompt_opt = teleprompter.compile(program, devset=devset[:DEV_NUM], optuna_trials_num=100, max_bootstrapped_demos=3, max_labeled_demos=5, eval_kwargs=kwargs)
# eval_score = evaluate(compiled_prompt_opt, devset=evalset[:EVAL_NUM], **kwargs)

# Note that this teleprompter takes in the following parameters:

# * prompt_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (ie. dspy.settings.configure(lm=task_model)).
# * task_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (ie. dspy.settings.configure(lm=task_model)).
# * metric: The task metric used for optimization.
# * n: The number of new prompts and sets of fewshot examples to generate and evaluate. Default=10.
# * init_temperature: The temperature used to generate new prompts. Higher roughly equals more creative. Default=1.0.
# * verbose: Tells the method whether or not to print intermediate steps.
# * track_stats: Tells the method whether or not to track statistics about the optimization process.
#                 If True, the method will track a dictionary with a key corresponding to the trial number, 
#                 and a value containing a dict with the following keys:
#                     * program: the program being evaluated at a given trial
#                     * score: the last average evaluated score for the program
#                     * pruned: whether or not this program was pruned
#                 This information will be returned as attributes of the best program.
# """


# class BayesianSignatureOptimizer(MIPRO):
#     def __init__(
#         self,
#         prompt_model=None,
#         task_model=None,
#         teacher_settings={},
#         n=10,
#         metric=None,
#         init_temperature=1.0,
#         verbose=False,
#         track_stats=True,
#         view_data_batch_size=10,
#     ):
#         print(
#             "\u001b[31m[WARNING] BayesianSignatureOptimizer has been deprecated and replaced with MIPRO.  BayesianSignatureOptimizer will be removed in a future release. \u001b[31m",
#         )

#         super().__init__(
#             metric=metric,
#             prompt_model=prompt_model,
#             task_model=task_model,
#             teacher_settings=teacher_settings,
#             num_candidates=n,
#             init_temperature=init_temperature,
#             verbose=verbose,
#             track_stats=track_stats,
#             view_data_batch_size=view_data_batch_size,
#         )

#     def compile(
#         self,
#         student,
#         *,
#         devset,
#         max_bootstrapped_demos,
#         max_labeled_demos,
#         eval_kwargs,
#         seed=42,
#         optuna_trials_num,
#         view_data=True,
#         view_examples=True,
#         requires_permission_to_run=False,
#         num_trials=None,
#     ):
#         return super().compile(
#             student,
#             trainset=devset,
#             max_bootstrapped_demos=max_bootstrapped_demos,
#             max_labeled_demos=max_labeled_demos,
#             eval_kwargs=eval_kwargs,
#             seed=seed,
#             view_data=view_data,
#             view_examples=view_examples,
#             requires_permission_to_run=requires_permission_to_run,
#             num_trials=optuna_trials_num,
#         )



import json
from typing import Any, List, Literal, Union, get_args, get_origin

from pydantic import TypeAdapter
from pydantic.fields import FieldInfo

from .image_utils import Image, encode_image


def serialize_for_json(value: Any) -> Any:
    """
    Formats the specified value so that it can be serialized as a JSON string.

    Args:
        value: The value to format as a JSON string.
    Returns:
        The formatted value, which is serializable as a JSON string.
    """
    # Attempt to format the value as a JSON-compatible object using pydantic, falling back to
    # a string representation of the value if that fails (e.g. if the value contains an object
    # that pydantic doesn't recognize or can't serialize)
    try:
        return TypeAdapter(type(value)).dump_python(value, mode="json")
    except Exception:
        return str(value)


def format_field_value(field_info: FieldInfo, value: Any, assume_text=True) -> Union[str, dict]:
    """
    Formats the value of the specified field according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself.

    Args:
      field_info: Information about the field, including its DSPy field type and annotation.
      value: The value of the field.
    Returns:
      The formatted value of the field, represented as a string.
    """
    string_value = None
    if isinstance(value, list) and field_info.annotation is str:
        # If the field has no special type requirements, format it as a nice numbered list for the LM.
        string_value = _format_input_list_field_value(value)
    else:
        jsonable_value = serialize_for_json(value)
        if isinstance(jsonable_value, dict) or isinstance(jsonable_value, list):
            string_value = json.dumps(jsonable_value, ensure_ascii=False)
        else:
            # If the value is not a Python representation of a JSON object or Array
            # (e.g. the value is a JSON string), just use the string representation of the value
            # to avoid double-quoting the JSON string (which would hurt accuracy for certain
            # tasks, e.g. tasks that rely on computing string length)
            string_value = str(jsonable_value)

    if assume_text:
        return string_value
    elif isinstance(value, Image) or field_info.annotation == Image:
        # This validation should happen somewhere else
        # Safe to import PIL here because it's only imported when an image is actually being formatted
        try:
            import PIL
        except ImportError:
            raise ImportError("PIL is required to format images; Run `pip install pillow` to install it.")
        image_value = value
        if not isinstance(image_value, Image):
            if isinstance(image_value, dict) and "url" in image_value:
                image_value = image_value["url"]
            elif isinstance(image_value, str):
                image_value = encode_image(image_value)
            elif isinstance(image_value, PIL.Image.Image):
                image_value = encode_image(image_value)
            assert isinstance(image_value, str)
            image_value = Image(url=image_value)
        return {"type": "image_url", "image_url": image_value.model_dump()}
    else:
        return {"type": "text", "text": string_value}


def find_enum_member(enum, identifier):
    """
    Finds the enum member corresponding to the specified identifier, which may be the
    enum member's name or value.

    Args:
        enum: The enum to search for the member.
        identifier: If the enum is explicitly-valued, this is the value of the enum member to find.
                    If the enum is auto-valued, this is the name of the enum member to find.
    Returns:
        The enum member corresponding to the specified identifier.
    """
    # Check if the identifier is a valid enum member value *before* checking if it's a valid enum
    # member name, since the identifier will be a value for explicitly-valued enums. This handles
    # the (rare) case where an enum member value is the same as another enum member's name in
    # an explicitly-valued enum
    for member in enum:
        if member.value == identifier:
            return member

    # If the identifier is not a valid enum member value, check if it's a valid enum member name,
    # since the identifier will be a member name for auto-valued enums
    if identifier in enum.__members__:
        return enum[identifier]

    raise ValueError(f"{identifier} is not a valid name or value for the enum {enum.__name__}")


def get_annotation_name(annotation):
    origin = get_origin(annotation)
    args = get_args(annotation)
    if origin is None:
        if hasattr(annotation, "__name__"):
            return annotation.__name__
        else:
            return str(annotation)

    if origin is Literal:
        args_str = ", ".join(
            _quoted_string_for_literal_type_annotation(a) if isinstance(a, str) else get_annotation_name(a)
            for a in args
        )
        return f"{get_annotation_name(origin)}[{args_str}]"
    else:
        args_str = ", ".join(get_annotation_name(a) for a in args)
        return f"{get_annotation_name(origin)}[{args_str}]"


def _format_input_list_field_value(value: List[Any]) -> str:
    """
    Formats the value of an input field of type List[Any].

    Args:
      value: The value of the list-type input field.
    Returns:
      A string representation of the input field's list value.
    """
    if len(value) == 0:
        return "N/A"
    if len(value) == 1:
        return _format_blob(value[0])

    return "\n".join([f"[{idx+1}] {_format_blob(txt)}" for idx, txt in enumerate(value)])


def _format_blob(blob: str) -> str:
    """
    Formats the specified text blobs so that an LM can parse it correctly within a list
    of multiple text blobs.

    Args:
        blob: The text blob to format.
    Returns:
        The formatted text blob.
    """
    if "\n" not in blob and "«" not in blob and "»" not in blob:
        return f"«{blob}»"

    modified_blob = blob.replace("\n", "\n    ")
    return f"«««\n    {modified_blob}\n»»»"


def _quoted_string_for_literal_type_annotation(s: str) -> str:
    """
    Return the specified string quoted for inclusion in a literal type annotation.
    """
    has_single = "'" in s
    has_double = '"' in s

    if has_single and not has_double:
        # Only single quotes => enclose in double quotes
        return f'"{s}"'
    elif has_double and not has_single:
        # Only double quotes => enclose in single quotes
        return f"'{s}'"
    elif has_single and has_double:
        # Both => enclose in single quotes; escape each single quote with \'
        escaped = s.replace("'", "\\'")
        return f"'{escaped}'"
    else:
        # Neither => enclose in single quotes
        return f"'{s}'"



# import functools
# import os
# from typing import List, Optional

# import openai

# import dspy
# from dspy.dsp.modules.cache_utils import NotebookCacheMemory, cache_turn_on
# from dspy.dsp.utils import dotdict

# # Check for necessary libraries and suggest installation if not found.
# try:
#     import clickhouse_connect
# except ImportError:
#     raise ImportError(
#         "The 'myscale' extra is required to use MyScaleRM. Install it with `pip install dspy-ai[myscale]`",
#     )

# # Verify the compatibility of the OpenAI library version installed.
# try:
#     major, minor, _ = map(int, openai.__version__.split("."))
#     OPENAI_VERSION_COMPATIBLE = major >= 1 and minor >= 16
# except Exception:
#     OPENAI_VERSION_COMPATIBLE = False

# if not OPENAI_VERSION_COMPATIBLE:
#     raise ImportError(
#         "An incompatible OpenAI library version is installed. Ensure you have version 1.16.1 or later.",
#     )

# # Attempt to handle specific OpenAI errors; fallback to general ones if necessary.
# try:
#     import openai.error

#     ERRORS = (openai.error.RateLimitError, openai.error.ServiceUnavailableError, openai.error.APIError)
# except Exception:
#     ERRORS = (openai.RateLimitError, openai.APIError)


# class MyScaleRM(dspy.Retrieve):
#     """
#     A retrieval module that uses MyScaleDB to return the top passages for a given query.

#     MyScaleDB is a fork of ClickHouse that focuses on vector similarity search and full
#     text search.  MyScaleRM is designed to facilitate easy retrieval of information from
#     MyScaleDB using embeddings.  It supports embedding generation through either a local
#     model or the OpenAI API. This class abstracts away the complexities of connecting to
#     MyScaleDB, managing API keys, and processing queries to return semantically
#     relevant results.

#     Assumes that a table named `database.table` exists in MyScaleDB, and that the
#     table has column named `vector_column` that stores vector data and a vector index has
#     been created on this column. Other metadata are stored in `metadata_columns`.

#     Args:
#         client (clickhouse_connect.driver.client.Client): A client connection to the MyScaleDB.
#         table (str): Name of the table within the database to perform queries against.
#         database (str, optional): Name of the database to query within MyScaleDB.
#         metadata_columns(List[str], optional): A list of columns to include in the results.
#         vector_column (str, optional): The name of the column in the table that stores vector data.
#         k (int, optional): The number of closest matches to retrieve for a given query.
#         openai_api_key (str, optional): The API key for accessing OpenAI's services.
#         model (str, optional): Specifies the particular OpenAI model to use for embedding generation.
#         use_local_model (bool): Flag indicating whether a local model is used for embeddings.

#     """

#     def __init__(
#         self,
#         client: clickhouse_connect.driver.client.Client,
#         table: str,
#         database: str = "default",
#         metadata_columns: List[str] = ["text"],
#         vector_column: str = "vector",
#         k: int = 3,
#         openai_api_key: Optional[str] = None,
#         openai_model: Optional[str] = None,
#         local_embed_model: Optional[str] = None,
#     ):
#         self.client = client
#         self.database = database
#         self.table = table
#         if not metadata_columns:
#             raise ValueError("metadata_columns is required")
#         self.metadata_columns = metadata_columns
#         self.vector_column = vector_column
#         self.k = k
#         self.openai_api_key = openai_api_key
#         self.model = openai_model
#         self.use_local_model = False

#         if local_embed_model:
#             self.setup_local_model(local_embed_model)
#         elif openai_api_key:
#             os.environ["OPENAI_API_KEY"] = self.openai_api_key

#     def setup_local_model(self, model_name: str):
#         """
#         Configures a local model for embedding generation, including model and tokenizer loading.

#         Args:
#             model_name: The name or path to the pre-trained model to load.

#         Raises:
#             ModuleNotFoundError: If necessary libraries (torch or transformers) are not installed.
#         """
#         try:
#             import torch
#             from transformers import AutoModel, AutoTokenizer
#         except ImportError as exc:
#             raise ModuleNotFoundError(
#                 """You need to install PyTorch and Hugging Face's transformers library to use a local embedding model.
#                 Install the pytorch using `pip install torch` and transformers using `pip install transformers` """,
#             ) from exc

#         try:
#             self._local_embed_model = AutoModel.from_pretrained(model_name)
#             self._local_tokenizer = AutoTokenizer.from_pretrained(model_name)
#             self.use_local_model = True
#         except Exception as e:
#             raise ValueError(f"Failed to load model or tokenizer. Error: {str(e)}")

#         if torch.cuda.is_available():
#             self.device = torch.device("cuda:0")
#         elif torch.backends.mps.is_available():
#             self.device = torch.device("mps")
#         else:
#             self.device = torch.device("cpu")

#         self._local_embed_model.to(self.device)

#     @functools.lru_cache(maxsize=None if cache_turn_on else 0)
#     @NotebookCacheMemory.cache
#     def get_embeddings(self, query: str) -> List[float]:
#         """
#         Determines the appropriate source (OpenAI or local model) for embedding generation based on class configuration,
#         and retrieves embeddings for the provided queries.

#         Args:
#             query: A query to generate embeddings for.

#         Returns:
#             A list of embeddings corresponding to the query in the input list.

#         Raises:
#             ValueError: If neither an OpenAI API key nor a local model has been configured.
#         """
#         if self.openai_api_key and self.model:
#             return self._get_embeddings_from_openai(query)
#         elif self.use_local_model:
#             return self._get_embedding_from_local_model(query)
#         else:
#             raise ValueError("No valid method for obtaining embeddings is configured.")

#     # TO DO Add this method as Util method outside MyScaleRM
#     def _get_embeddings_from_openai(self, query: str) -> List[float]:
#         """
#         Uses the OpenAI API to generate embeddings for the given query.

#         Args:
#             query: A string for which to generate embeddings.

#         Returns:
#             A list containing the embedding of a query.
#         """
#         response = openai.embeddings.create(model=self.model, input=query)
#         return response.data[0].embedding

#     # TO DO Add this method as Util method outside MyScaleRM
#     def _get_embedding_from_local_model(self, query: str) -> List[float]:
#         """
#         Generates embeddings for a single query using the configured local model.

#         Args:
#             query: The text query to generate an embedding for.

#         Returns:
#             A list representing the query's embedding.
#         """
#         import torch

#         self._local_embed_model.eval()  # Ensure the model is in evaluation mode

#         inputs = self._local_tokenizer(query, return_tensors="pt", padding=True, truncation=True).to(self.device)
#         with torch.no_grad():
#             output = self._local_embed_model(**inputs)

#         return output.last_hidden_state.mean(dim=1).squeeze().numpy().tolist()

#     def forward(self, user_query: str, k: Optional[int] = None) -> List[dotdict]:
#         """
#         Executes a retrieval operation based on a user's query and returns the top k relevant results.

#         Args:
#             user_query: The query text to search for.
#             k: Optional; The number of top matches to return. Defaults to the class's configured k value.

#         Returns:
#             A list of dotdict objects containing the formatted retrieval results.

#         Raises:
#             ValueError: If the user_query is None.
#         """
#         if user_query is None:
#             raise ValueError("Query is required")
#         k = k or self.k
#         embeddings = self.get_embeddings(user_query)
#         columns_string = ", ".join(self.metadata_columns)
#         result = self.client.query(f"""
#         SELECT {columns_string},
#         distance({self.vector_column}, {embeddings}) as dist FROM {self.database}.{self.table} ORDER BY dist LIMIT {k}
#         """)
#         result = self.client.query(f"""
#             SELECT {columns_string}, distance({self.vector_column}, {embeddings}) AS dist
#             FROM {self.database}.{self.table}
#             ORDER BY dist
#             LIMIT {k}
#         """)

#         # Convert the metadata into strings to pass to dspy.Prediction
#         results = []
#         for row in result.named_results():
#             if len(self.metadata_columns) == 1:
#                 results.append(row[self.metadata_columns[0]])
#             else:
#                 row_strings = [f"{column}: {row[column]}" for column in self.metadata_columns]  # Format row data
#                 row_string = "\n".join(row_strings)  # Combine formatted data
#                 results.append(row_string)  # Append to results

#         # Return list of dotdict
#         return [dotdict({"long_text": passage}) for passage in results]



import logging
import types
from typing import Any, Callable, List, Optional

import pandas as pd
import tqdm

import dspy
from dspy.utils.parallelizer import ParallelExecutor

try:
    from IPython.display import HTML
    from IPython.display import display as display

except ImportError:

    def display(obj: Any):
        """
        Display the specified Python object in the console.

        :param obj: The Python object to display.
        """
        print(obj)

    def HTML(x: str) -> str:
        """
        Obtain the HTML representation of the specified string.
        """
        # NB: This method exists purely for code compatibility with the IPython HTML() function in
        # environments where IPython is not available. In such environments where IPython is not
        # available, this method will simply return the input string.
        return x


# TODO: Counting failures and having a max_failure count. When that is exceeded (also just at the end),
# we print the number of failures, the first N examples that failed, and the first N exceptions raised.

logger = logging.getLogger(__name__)


class Evaluate:
    """DSPy Evaluate class.

    This class is used to evaluate the performance of a DSPy program. Users need to provide a evaluation dataset and 
    a metric function in order to use this class. This class supports parallel evaluation on the provided dataset.
    """
    def __init__(
        self,
        *,
        devset: List["dspy.Example"],
        metric: Optional[Callable] = None,
        num_threads: int = 1,
        display_progress: bool = False,
        display_table: bool = False,
        max_errors: int = 5,
        return_all_scores: bool = False,
        return_outputs: bool = False,
        provide_traceback: bool = False,
        failure_score: float = 0.0,
        **kwargs,
    ):
        """
        Args:
            devset (List[dspy.Example]): the evaluation dataset.
            metric (Callable): The metric function to use for evaluation.
            num_threads (int): The number of threads to use for parallel evaluation.
            display_progress (bool): Whether to display progress during evaluation.
            display_table (bool): Whether to display the evaluation results in a table.
            max_errors (int): The maximum number of errors to allow before stopping evaluation.
            return_all_scores (bool): Whether to return scores for every data record in `devset`.
            return_outputs (bool): Whether to return the dspy program's outputs for every data in `devset`.
            provide_traceback (bool): Whether to provide traceback information during evaluation.
            failure_score (float): The default score to use if evaluation fails due to an exception.
        """
        self.devset = devset
        self.metric = metric
        self.num_threads = num_threads
        self.display_progress = display_progress
        self.display_table = display_table
        self.max_errors = max_errors
        self.return_all_scores = return_all_scores
        self.return_outputs = return_outputs
        self.provide_traceback = provide_traceback
        self.failure_score = failure_score

    def __call__(
        self,
        program: "dspy.Module",
        metric: Optional[Callable] = None,
        devset: Optional[List["dspy.Example"]] = None,
        num_threads: Optional[int] = None,
        display_progress: Optional[bool] = None,
        display_table: Optional[bool] = None,
        return_all_scores: Optional[bool] = None,
        return_outputs: Optional[bool] = None,
    ):
        """
        Args:
            program (dspy.Module): The DSPy program to evaluate.
            metric (Callable): The metric function to use for evaluation. if not provided, use `self.metric`.
            devset (List[dspy.Example]): the evaluation dataset. if not provided, use `self.devset`.
            num_threads (int): The number of threads to use for parallel evaluation. if not provided, use
                `self.num_threads`.
            display_progress (bool): Whether to display progress during evaluation. if not provided, use
                `self.display_progress`.
            display_table (bool): Whether to display the evaluation results in a table. if not provided, use
                `self.display_table`.
            return_all_scores (bool): Whether to return scores for every data record in `devset`. if not provided,
                use `self.return_all_scores`.
            return_outputs (bool): Whether to return the dspy program's outputs for every data in `devset`. if not
                provided, use `self.return_outputs`.

        Returns:
            The evaluation results are returned in different formats based on the flags:
            
            - Base return: A float percentage score (e.g., 67.30) representing overall performance
            
            - With `return_all_scores=True`:
                Returns (overall_score, individual_scores) where individual_scores is a list of 
                float scores for each example in devset
            
            - With `return_outputs=True`:
                Returns (overall_score, result_triples) where result_triples is a list of 
                (example, prediction, score) tuples for each example in devset

            - With both flags=True:
                Returns (overall_score, result_triples, individual_scores)

        """
        metric = metric if metric is not None else self.metric
        devset = devset if devset is not None else self.devset
        num_threads = num_threads if num_threads is not None else self.num_threads
        display_progress = display_progress if display_progress is not None else self.display_progress
        display_table = display_table if display_table is not None else self.display_table
        return_all_scores = return_all_scores if return_all_scores is not None else self.return_all_scores
        return_outputs = return_outputs if return_outputs is not None else self.return_outputs

        tqdm.tqdm._instances.clear()

        executor = ParallelExecutor(
            num_threads=num_threads,
            disable_progress_bar=not display_progress,
            max_errors=self.max_errors,
            provide_traceback=self.provide_traceback,
            compare_results=True,
        )

        def process_item(example):
            prediction = program(**example.inputs())
            score = metric(example, prediction)

            # Increment assert and suggest failures to program's attributes
            if hasattr(program, "_assert_failures"):
                program._assert_failures += dspy.settings.get("assert_failures")
            if hasattr(program, "_suggest_failures"):
                program._suggest_failures += dspy.settings.get("suggest_failures")

            return prediction, score

        results = executor.execute(process_item, devset)
        assert len(devset) == len(results)

        results = [((dspy.Prediction(), self.failure_score) if r is None else r) for r in results]
        results = [(example, prediction, score) for example, (prediction, score) in zip(devset, results)]
        ncorrect, ntotal = sum(score for *_, score in results), len(devset)

        logger.info(f"Average Metric: {ncorrect} / {ntotal} ({round(100 * ncorrect / ntotal, 1)}%)")
            
        def prediction_is_dictlike(prediction):
            # Downstream logic for displaying dictionary-like predictions depends solely on the predictions
            # having a method called `items()` for iterating through key/value pairs
            return hasattr(prediction, "items") and callable(getattr(prediction, "items"))

        data = [
            (
                merge_dicts(example, prediction) | {"correct": score}
                if prediction_is_dictlike(prediction)
                else dict(example) | {"prediction": prediction, "correct": score}
            )
            for example, prediction, score in results
        ]


        # Truncate every cell in the DataFrame (DataFrame.applymap was renamed to DataFrame.map in Pandas 2.1.0)
        result_df = pd.DataFrame(data)
        result_df = result_df.map(truncate_cell) if hasattr(result_df, "map") else result_df.applymap(truncate_cell)

        # Rename the 'correct' column to the name of the metric object
        metric_name = metric.__name__ if isinstance(metric, types.FunctionType) else metric.__class__.__name__
        result_df = result_df.rename(columns={"correct": metric_name})

        if display_table:
            if isinstance(display_table, bool):
                df_to_display = result_df.copy()
                truncated_rows = 0
            else:
                df_to_display = result_df.head(display_table).copy()
                truncated_rows = len(result_df) - display_table

            df_to_display = stylize_metric_name(df_to_display, metric_name)

            display_dataframe(df_to_display)

            if truncated_rows > 0:
                # Simplified message about the truncated rows
                message = f"""
                <div style='
                    text-align: center;
                    font-size: 16px;
                    font-weight: bold;
                    color: #555;
                    margin: 10px 0;'>
                    ... {truncated_rows} more rows not displayed ...
                </div>
                """
                display(HTML(message))

        if return_all_scores and return_outputs:
            return round(100 * ncorrect / ntotal, 2), results, [score for *_, score in results]
        if return_all_scores:
            return round(100 * ncorrect / ntotal, 2), [score for *_, score in results]
        if return_outputs:
            return round(100 * ncorrect / ntotal, 2), results

        return round(100 * ncorrect / ntotal, 2)


def merge_dicts(d1, d2) -> dict:
    merged = {}
    for k, v in d1.items():
        if k in d2:
            merged[f"example_{k}"] = v
        else:
            merged[k] = v

    for k, v in d2.items():
        if k in d1:
            merged[f"pred_{k}"] = v
        else:
            merged[k] = v

    return merged


def truncate_cell(content) -> str:
    """Truncate content of a cell to 25 words."""
    words = str(content).split()
    if len(words) > 25:
        return " ".join(words[:25]) + "..."
    return content


def stylize_metric_name(df: pd.DataFrame, metric_name: str) -> pd.DataFrame:
    """
    Stylize the cell contents of a pandas DataFrame corresponding to the specified metric name.

    :param df: The pandas DataFrame for which to stylize cell contents.
    :param metric_name: The name of the metric for which to stylize DataFrame cell contents.
    """
    df[metric_name] = df[metric_name].apply(
        lambda x: f"✔️ [{x:.3f}]" if x and isinstance(x, float) else f"✔️ [{x}]" if x else ""
    )
    return df


def display_dataframe(df: pd.DataFrame):
    """
    Display the specified Pandas DataFrame in the console.

    :param df: The Pandas DataFrame to display.
    """
    if is_in_ipython_notebook_environment():
        display(configure_dataframe_for_ipython_notebook_display(df))
    else:
        # Pretty print the DataFrame to the console
        with pd.option_context(
            "display.max_rows", None, "display.max_columns", None
        ):  # more options can be specified also
            print(df)


def configure_dataframe_for_ipython_notebook_display(df: pd.DataFrame) -> pd.DataFrame:
    """Set various pandas display options for DataFrame in an IPython notebook environment."""
    pd.options.display.max_colwidth = 70
    return df


def is_in_ipython_notebook_environment():
    """
    Check if the current environment is an IPython notebook environment.

    :return: True if the current environment is an IPython notebook environment, False otherwise.
    """
    try:
        from IPython import get_ipython

        # This is a best-effort check to see if we are in an IPython notebook environment
        return "IPKernelApp" in getattr(get_ipython(), "config", {})
    except ImportError:
        return False


# FIXME: TODO: The merge_dicts stuff above is way too quick and dirty.
# TODO: the display_table can't handle False but can handle 0!
# Not sure how it works with True exactly, probably fails too.



import dspy

from tqdm import tqdm
from copy import deepcopy
from random import sample
from pydantic import BaseModel
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Tuple, Optional

from .teleprompt import Teleprompter
from dspy.predict.avatar import ActionOutput


DEFAULT_MAX_EXAMPLES = 10


class EvalResult(BaseModel):
    example: dict
    score: float
    actions: Optional[List[ActionOutput]] = None


class Comparator(dspy.Signature):
    """After executing the given actions on user inputs using the given instruction, some inputs have yielded good, results, while others have not. I'll provide you the inputs along with their, corresponding evaluation metrics:

Task:
(1) Firstly, identify and contrast the patterns of inputs that have achieved good results with those that have not.
(2) Then, review the computational logic for any inconsistencies in the previous actions.
(3) Lastly, specify the modification in tools used that can lead to improved performance on the negative inputs."""

    instruction: str = dspy.InputField(
        prefix="Instruction: ",
        desc="Instruction for the actor to execute the task",
    )
    actions: List[str] = dspy.InputField(
        prefix="Actions: ",
        desc="Actions actor can take to complete the task",
    )
    pos_input_with_metrics: List[EvalResult] = dspy.InputField(
        prefix="Positive Inputs: ",
        desc="Positive inputs along with their score on a evaluation metric and actions taken",
    )
    neg_input_with_metrics: List[EvalResult] = dspy.InputField(
        prefix="Negative Inputs: ",
        desc="Negative inputs along with their score on a evaluation metric and actions taken",
    )
    feedback: str = dspy.OutputField(
        prefix="Feedback: ",
        desc="Feedback for the actor to improve the performance of negative inputs",
    )


class FeedbackBasedInstruction(dspy.Signature):
    """There is a task that needs to be completed for which one can use multiple tools to achieve the desired outcome. A group's performance was evaluated on a dataset of inputs, the inputs that did well are positive inputs, and the inputs that did not do well are negative inputs.

You received feedback on how they can better use the tools to improve your performance on the negative inputs. You have been provided with the previous instruction, that they followed to use tools to complete the task, and the feedback on your performance.

Your task is to incorporate the feedback and generate a detailed instruction for the group to follow to improve their performance on the task.

Make sure that the new instruction talks about how to use the tools effectively and should be no more than 3 paragraphs long. The previous instruction contains general guidelines that you must retain in the new instruction."""

    previous_instruction: str = dspy.InputField(
        prefix="Previous Instruction: ",
        desc="Previous instruction for the actor to execute the task",
    )
    feedback: str = dspy.InputField(
        prefix="Feedback: ",
        desc="Feedback for the actor to improve the performance of negative inputs",
    )
    new_instruction: str = dspy.OutputField(
        prefix="New Instruction: ",
        desc="New instruction for the actor to execute the task",
    )


class AvatarOptimizer(Teleprompter):
    def __init__(
        self,
        metric: Callable,
        max_iters: int = 10,
        lower_bound: int = 0,
        upper_bound: int = 1,
        max_positive_inputs: int = None,
        max_negative_inputs: int = None,
        optimize_for: str = "max",
    ):
        assert metric is not None, "`metric` argument cannot be None. Please provide a metric function."
        self.metric = metric
        self.optimize_for = optimize_for

        self.max_iters = max_iters

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        self.max_positive_inputs = max_positive_inputs or DEFAULT_MAX_EXAMPLES
        self.max_negative_inputs = max_negative_inputs or DEFAULT_MAX_EXAMPLES

        self.comparator = dspy.TypedPredictor(Comparator)
        self.feedback_instruction = dspy.Predict(FeedbackBasedInstruction)

    def process_example(self, actor, example, return_outputs):
        actor = deepcopy(actor)

        try:
            prediction = actor(**example.inputs().toDict())
            score = self.metric(example, prediction)

            if return_outputs:
                return example, prediction, score
            else:
                return score

        except Exception as e:
            print(e)
            
            if return_outputs:
                return example, None, 0
            else:
                return 0


    def thread_safe_evaluator(self, devset, actor, return_outputs=False, num_threads=60):
        total_score = 0
        total_examples = len(devset)
        results = []

        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = [executor.submit(self.process_example, actor, example, return_outputs) for example in devset]
            
            for future in tqdm(futures, total=total_examples, desc="Processing examples"):
                result = future.result()
                if return_outputs:
                    example, prediction, score = result
                    total_score += score
                    results.append((example, prediction, score))
                else:
                    total_score += result
        
        avg_metric = total_score / total_examples
        
        if return_outputs:
            return avg_metric, results
        else:
            return avg_metric


    def _get_pos_neg_results(
        self, 
        actor: dspy.Module, 
        trainset: List[dspy.Example]
    ) -> Tuple[float, List[EvalResult], List[EvalResult]]:
        pos_inputs = []
        neg_inputs = []
        
        avg_score, results = self.thread_safe_evaluator(trainset, actor, return_outputs=True)
        print(f"Average Score: {avg_score}")

        for example, prediction, score in results:
            if score >= self.upper_bound:
                pos_inputs.append(
                    EvalResult(
                        example=example.inputs().toDict(),
                        score=score,
                        actions=prediction.actions if prediction else None
                    )
                )
            elif score <= self.lower_bound:
                neg_inputs.append(
                    EvalResult(
                        example=example.inputs().toDict(),
                        score=score,
                        actions=prediction.actions if prediction else None
                    )
                )

        if len(pos_inputs) == 0:
            raise ValueError("No positive examples found, try lowering the upper_bound or providing more training data")
        if len(neg_inputs) == 0:
            raise ValueError("No negative examples found, try raising the lower_bound or providing more training data")
        
        return (avg_score, pos_inputs, neg_inputs)
    

    def compile(self, student, *, trainset):
        best_actor = deepcopy(student)
        best_score = -999 if self.optimize_for == "max" else 999
        
        for i in range(self.max_iters):
            print(20*'=')
            print(f"Iteration {i+1}/{self.max_iters}")

            score, pos_inputs, neg_inputs = self._get_pos_neg_results(best_actor, trainset)
            print(f"Positive examples: {len(pos_inputs)}")
            print(f"Negative examples: {len(neg_inputs)}")
            print(f"Sampling {self.max_positive_inputs} positive examples and {self.max_negative_inputs} negative examples")

            if self.max_positive_inputs and len(pos_inputs) > self.max_positive_inputs:
                pos_inputs = sample(pos_inputs, self.max_positive_inputs)

            if self.max_negative_inputs and len(neg_inputs) > self.max_negative_inputs:
                neg_inputs = sample(neg_inputs, self.max_negative_inputs)

            feedback = self.comparator(
                instruction=best_actor.actor.signature.instructions,
                actions=[str(tool) for tool in best_actor.tools],
                pos_input_with_metrics=pos_inputs,
                neg_input_with_metrics=neg_inputs
            ).feedback

            new_instruction = self.feedback_instruction(
                previous_instruction=best_actor.actor.signature.instructions,
                feedback=feedback
            ).new_instruction

            print(f"Generated new instruction: {new_instruction}")

            if (self.optimize_for == "max" and best_score < score) or (self.optimize_for == "min" and best_score > score):
                best_actor.actor.signature = best_actor.actor.signature.with_instructions(new_instruction)
                best_actor.actor_clone = deepcopy(best_actor.actor)
                best_score = score
        
        print(f"Best Actor: {best_actor}")

        return best_actor



import ast
import enum
import inspect
import json
import re
import textwrap
from collections.abc import Mapping
from itertools import chain
from typing import Any, Dict, List, Literal, NamedTuple, Union

import pydantic
from pydantic import TypeAdapter
from pydantic.fields import FieldInfo

from dspy.adapters.base import Adapter
from dspy.adapters.utils import find_enum_member, format_field_value, get_annotation_name
from dspy.signatures.field import OutputField
from dspy.signatures.signature import Signature, SignatureMeta
from dspy.signatures.utils import get_dspy_field_type

field_header_pattern = re.compile(r"\[\[ ## (\w+) ## \]\]")


class FieldInfoWithName(NamedTuple):
    name: str
    info: FieldInfo


# Built-in field indicating that a chat turn has been completed.
BuiltInCompletedOutputFieldInfo = FieldInfoWithName(name="completed", info=OutputField())


class ChatAdapter(Adapter):
    def format(self, signature: Signature, demos: list[dict[str, Any]], inputs: dict[str, Any]) -> list[dict[str, Any]]:
        messages: list[dict[str, Any]] = []

        # Extract demos where some of the output_fields are not filled in.
        incomplete_demos = [
            demo for demo in demos if not all(k in demo and demo[k] is not None for k in signature.fields)
        ]
        complete_demos = [demo for demo in demos if demo not in incomplete_demos]
        # Filter out demos that don't have at least one input and one output field.
        incomplete_demos = [
            demo
            for demo in incomplete_demos
            if any(k in demo for k in signature.input_fields) and any(k in demo for k in signature.output_fields)
        ]

        demos = incomplete_demos + complete_demos

        prepared_instructions = prepare_instructions(signature)
        messages.append({"role": "system", "content": prepared_instructions})

        for demo in demos:
            messages.append(format_turn(signature, demo, role="user", incomplete=demo in incomplete_demos))
            messages.append(format_turn(signature, demo, role="assistant", incomplete=demo in incomplete_demos))

        messages.append(format_turn(signature, inputs, role="user"))
        return messages

    def parse(self, signature, completion):
        sections = [(None, [])]

        for line in completion.splitlines():
            match = field_header_pattern.match(line.strip())
            if match:
                sections.append((match.group(1), []))
            else:
                sections[-1][1].append(line)

        sections = [(k, "\n".join(v).strip()) for k, v in sections]

        fields = {}
        for k, v in sections:
            if (k not in fields) and (k in signature.output_fields):
                try:
                    fields[k] = parse_value(v, signature.output_fields[k].annotation)
                except Exception as e:
                    raise ValueError(
                        f"Error parsing field {k}: {e}.\n\n\t\tOn attempting to parse the value\n```\n{v}\n```"
                    )

        if fields.keys() != signature.output_fields.keys():
            raise ValueError(f"Expected {signature.output_fields.keys()} but got {fields.keys()}")

        return fields

    # TODO(PR): Looks ok?
    def format_finetune_data(self, signature, demos, inputs, outputs):
        # Get system + user messages
        messages = self.format(signature, demos, inputs)

        # Add the assistant message
        role = "assistant"
        incomplete = False
        assistant_message = format_turn(signature, outputs, role, incomplete)
        messages.append(assistant_message)

        # Wrap the messages in a dictionary with a "messages" key
        return dict(messages=messages)

    def format_turn(self, signature, values, role, incomplete=False):
        return format_turn(signature, values, role, incomplete)

    def format_fields(self, signature, values, role):
        fields_with_values = {
            FieldInfoWithName(name=field_name, info=field_info): values.get(
                field_name, "Not supplied for this particular example."
            )
            for field_name, field_info in signature.fields.items()
            if field_name in values
        }

        return format_fields(fields_with_values)


def format_fields(fields_with_values: Dict[FieldInfoWithName, Any], assume_text=True) -> Union[str, List[dict]]:
    """
    Formats the values of the specified fields according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself. Joins the formatted values
    into a single string, which is is a multiline string if there are multiple fields.

    Args:
      fields_with_values: A dictionary mapping information about a field to its corresponding
                          value.
    Returns:
      The joined formatted values of the fields, represented as a string or a list of dicts
    """
    output = []
    for field, field_value in fields_with_values.items():
        formatted_field_value = format_field_value(field_info=field.info, value=field_value, assume_text=assume_text)
        if assume_text:
            output.append(f"[[ ## {field.name} ## ]]\n{formatted_field_value}")
        else:
            output.append({"type": "text", "text": f"[[ ## {field.name} ## ]]\n"})
            if isinstance(formatted_field_value, dict) and formatted_field_value.get("type") == "image_url":
                output.append(formatted_field_value)
            else:
                output[-1]["text"] += formatted_field_value["text"]
    if assume_text:
        return "\n\n".join(output).strip()
    else:
        return output


def parse_value(value, annotation):
    if annotation is str:
        return str(value)

    parsed_value = value

    if isinstance(annotation, enum.EnumMeta):
        return find_enum_member(annotation, value)
    elif isinstance(value, str):
        try:
            parsed_value = json.loads(value)
        except json.JSONDecodeError:
            try:
                parsed_value = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                parsed_value = value

    return TypeAdapter(annotation).validate_python(parsed_value)


def format_turn(signature, values, role, incomplete=False):
    """
    Constructs a new message ("turn") to append to a chat thread. The message is carefully formatted
    so that it can instruct an LLM to generate responses conforming to the specified DSPy signature.

    Args:
        signature: The DSPy signature to which future LLM responses should conform.
        values: A dictionary mapping field names (from the DSPy signature) to corresponding values
            that should be included in the message.
        role: The role of the message, which can be either "user" or "assistant".
        incomplete: If True, indicates that output field values are present in the set of specified
            ``values``. If False, indicates that ``values`` only contains input field values.

    Returns:
        A chat message that can be appended to a chat thread. The message contains two string fields:
        ``role`` ("user" or "assistant") and ``content`` (the message text).
    """
    fields_to_collapse = []
    content = []

    if role == "user":
        fields = signature.input_fields
        if incomplete:
            fields_to_collapse.append(
                {
                    "type": "text",
                    "text": "This is an example of the task, though some input or output fields are not supplied.",
                }
            )
    else:
        fields = signature.output_fields
        # Add the built-in field indicating that the chat turn has been completed
        fields[BuiltInCompletedOutputFieldInfo.name] = BuiltInCompletedOutputFieldInfo.info
        values = {**values, BuiltInCompletedOutputFieldInfo.name: ""}
    field_names = fields.keys()
    if not incomplete:
        if not set(values).issuperset(set(field_names)):
            raise ValueError(f"Expected {field_names} but got {values.keys()}")

    fields_to_collapse.extend(
        format_fields(
            fields_with_values={
                FieldInfoWithName(name=field_name, info=field_info): values.get(
                    field_name, "Not supplied for this particular example."
                )
                for field_name, field_info in fields.items()
            },
            assume_text=False,
        )
    )

    if role == "user":
        output_fields = list(signature.output_fields.keys())

        def type_info(v):
            return (
                f" (must be formatted as a valid Python {get_annotation_name(v.annotation)})"
                if v.annotation is not str
                else ""
            )

        if output_fields:
            fields_to_collapse.append(
                {
                    "type": "text",
                    "text": "Respond with the corresponding output fields, starting with the field "
                    + ", then ".join(f"`[[ ## {f} ## ]]`{type_info(v)}" for f, v in signature.output_fields.items())
                    + ", and then ending with the marker for `[[ ## completed ## ]]`.",
                }
            )

    # flatmap the list if any items are lists otherwise keep the item
    flattened_list = list(
        chain.from_iterable(item if isinstance(item, list) else [item] for item in fields_to_collapse)
    )

    if all(message.get("type", None) == "text" for message in flattened_list):
        content = "\n\n".join(message.get("text") for message in flattened_list)
        return {"role": role, "content": content}

    # Collapse all consecutive text messages into a single message.
    collapsed_messages = []
    for item in flattened_list:
        # First item is always added
        if not collapsed_messages:
            collapsed_messages.append(item)
            continue

        # If the current item is image, add to collapsed_messages
        if item.get("type") == "image_url":
            if collapsed_messages[-1].get("type") == "text":
                collapsed_messages[-1]["text"] += "\n"
            collapsed_messages.append(item)
        # If the previous item is text and current item is text, append to the previous item
        elif collapsed_messages[-1].get("type") == "text":
            collapsed_messages[-1]["text"] += "\n\n" + item["text"]
        # If the previous item is not text(aka image), add the current item as a new item
        else:
            item["text"] = "\n\n" + item["text"]
            collapsed_messages.append(item)

    return {"role": role, "content": collapsed_messages}


def enumerate_fields(fields: dict) -> str:
    parts = []
    for idx, (k, v) in enumerate(fields.items()):
        parts.append(f"{idx+1}. `{k}`")
        parts[-1] += f" ({get_annotation_name(v.annotation)})"
        parts[-1] += f": {v.json_schema_extra['desc']}" if v.json_schema_extra["desc"] != f"${{{k}}}" else ""

    return "\n".join(parts).strip()


def move_type_to_front(d):
    # Move the 'type' key to the front of the dictionary, recursively, for LLM readability/adherence.
    if isinstance(d, Mapping):
        return {k: move_type_to_front(v) for k, v in sorted(d.items(), key=lambda item: (item[0] != "type", item[0]))}
    elif isinstance(d, list):
        return [move_type_to_front(item) for item in d]
    return d


def prepare_schema(type_):
    schema = pydantic.TypeAdapter(type_).json_schema()
    schema = move_type_to_front(schema)
    return schema


def prepare_instructions(signature: SignatureMeta):
    parts = []
    parts.append("Your input fields are:\n" + enumerate_fields(signature.input_fields))
    parts.append("Your output fields are:\n" + enumerate_fields(signature.output_fields))
    parts.append("All interactions will be structured in the following way, with the appropriate values filled in.")

    def field_metadata(field_name, field_info):
        field_type = field_info.annotation

        if get_dspy_field_type(field_info) == "input" or field_type is str:
            desc = ""
        elif field_type is bool:
            desc = "must be True or False"
        elif field_type in (int, float):
            desc = f"must be a single {field_type.__name__} value"
        elif inspect.isclass(field_type) and issubclass(field_type, enum.Enum):
            desc = f"must be one of: {'; '.join(field_type.__members__)}"
        elif hasattr(field_type, "__origin__") and field_type.__origin__ is Literal:
            desc = (
                # Strongly encourage the LM to avoid choosing values that don't appear in the
                # literal or returning a value of the form 'Literal[<selected_value>]'
                f"must exactly match (no extra characters) one of: {'; '.join([str(x) for x in field_type.__args__])}"
            )
        else:
            desc = "must be pareseable according to the following JSON schema: "
            desc += json.dumps(prepare_schema(field_type), ensure_ascii=False)

        desc = (" " * 8) + f"# note: the value you produce {desc}" if desc else ""
        return f"{{{field_name}}}{desc}"

    def format_signature_fields_for_instructions(fields: Dict[str, FieldInfo]):
        return format_fields(
            fields_with_values={
                FieldInfoWithName(name=field_name, info=field_info): field_metadata(field_name, field_info)
                for field_name, field_info in fields.items()
            },
            assume_text=True,
        )

    parts.append(format_signature_fields_for_instructions(signature.input_fields))
    parts.append(format_signature_fields_for_instructions(signature.output_fields))
    parts.append(format_fields({BuiltInCompletedOutputFieldInfo: ""}, assume_text=True))
    instructions = textwrap.dedent(signature.instructions)
    objective = ("\n" + " " * 8).join([""] + instructions.splitlines())
    parts.append(f"In adhering to this structure, your objective is: {objective}")

    return "\n\n".join(parts).strip()



# from collections import defaultdict
# from typing import Optional, Union

# import dspy
# from dspy.dsp.modules.sentence_vectorizer import BaseSentenceVectorizer, FastEmbedVectorizer
# from dspy.dsp.utils import dotdict

# try:
#     from qdrant_client import QdrantClient, models
# except ImportError as e:
#     raise ImportError(
#         "The 'qdrant' extra is required to use QdrantRM. Install it with `pip install dspy-ai[qdrant]`",
#     ) from e


# class QdrantRM(dspy.Retrieve):
#     """A retrieval module that uses Qdrant to return the top passages for a given query.

#     Args:
#         qdrant_collection_name (str): The name of the Qdrant collection.
#         qdrant_client (QdrantClient): An instance of `qdrant_client.QdrantClient`.
#         k (int, optional): The default number of top passages to retrieve. Default: 3.
#         document_field (str, optional): The key in the Qdrant payload with the content. Default: `"document"`.
#         vectorizer (BaseSentenceVectorizer, optional): An implementation `BaseSentenceVectorizer`.
#                                                            Default: `FastEmbedVectorizer`.
#         vector_name (str, optional): Name of the vector in the collection. Default: The first available vector name.

#     Examples:
#         Below is a code snippet that shows how to use Qdrant as the default retriver:
#         ```python
#         from qdrant_client import QdrantClient

#         llm = dspy.OpenAI(model="gpt-3.5-turbo")
#         qdrant_client = QdrantClient()
#         retriever_model = QdrantRM("my_collection_name", qdrant_client=qdrant_client)
#         dspy.settings.configure(lm=llm, rm=retriever_model)
#         ```

#         Below is a code snippet that shows how to use Qdrant in the forward() function of a module
#         ```python
#         self.retrieve = QdrantRM(question, k=num_passages, filter=filter)
#         ```
#     """

#     def __init__(
#         self,
#         qdrant_collection_name: str,
#         qdrant_client: QdrantClient,
#         k: int = 3,
#         document_field: str = "document",
#         vectorizer: Optional[BaseSentenceVectorizer] = None,
#         vector_name: Optional[str] = None,
#     ):
#         self._collection_name = qdrant_collection_name
#         self._client = qdrant_client

#         self._vectorizer = vectorizer or FastEmbedVectorizer(self._client.embedding_model_name)

#         self._document_field = document_field

#         self._vector_name = vector_name or self._get_first_vector_name()

#         super().__init__(k=k)

#     def forward(self, query_or_queries: Union[str, list[str]], k: Optional[int] = None, filter: Optional[models.Filter]=None) -> dspy.Prediction:
#         """Search with Qdrant for self.k top passages for query.

#         Args:
#             query_or_queries (Union[str, List[str]]): The query or queries to search for.
#             k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.
#             filter (Optional["Filter"]): "Look only for points which satisfies this conditions". Default: None.

#         Returns:
#             dspy.Prediction: An object containing the retrieved passages.
#         """
#         queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
#         queries = [q for q in queries if q]  # Filter empty queries

#         vectors = self._vectorizer(queries)

#         search_requests = [
#             models.QueryRequest(
#                 query=vector,
#                 using=self._vector_name,
#                 limit=k or self.k,
#                 with_payload=[self._document_field],
#                 filter=filter,
#             )
#             for vector in vectors
#         ]
#         batch_results = self._client.query_batch_points(self._collection_name, requests=search_requests)

#         passages_scores = defaultdict(float)
#         for batch in batch_results:
#             for result in batch.points:
#                 # If a passage is returned multiple times, the score is accumulated.
#                 document = result.payload.get(self._document_field)
#                 passages_scores[document] += result.score

#         # Sort passages by their accumulated scores in descending order
#         sorted_passages = sorted(passages_scores.items(), key=lambda x: x[1], reverse=True)[:k]

#         # Wrap each sorted passage in a dotdict with 'long_text'
#         return [dotdict({"long_text": passage}) for passage, _ in sorted_passages]

#     def _get_first_vector_name(self) -> Optional[str]:
#         vectors = self._client.get_collection(self._collection_name).config.params.vectors

#         if not isinstance(vectors, dict):
#             # The collection only has the default, unnamed vector
#             return None

#         first_vector_name = list(vectors.keys())[0]

#         # The collection has multiple vectors. Could also include the falsy unnamed vector - Empty string("")
#         return first_vector_name or None



from dspy.dsp.utils import EM, normalize_text

from dspy.evaluate.metrics import answer_exact_match, answer_passage_match
from dspy.evaluate.evaluate import Evaluate
from dspy.evaluate.auto_evaluation import SemanticF1, CompleteAndGrounded

__all__ = [
    "EM",
    "normalize_text",
    "answer_exact_match",
    "answer_passage_match",
    "Evaluate",
    "SemanticF1",
    "CompleteAndGrounded",
]



import ast
import enum
import inspect
import json
import logging
import textwrap
from copy import deepcopy
from typing import Any, Dict, KeysView, Literal, NamedTuple

import json_repair
import litellm
import pydantic
from pydantic import TypeAdapter, create_model
from pydantic.fields import FieldInfo

from dspy.adapters.base import Adapter
from dspy.adapters.image_utils import Image
from dspy.adapters.utils import find_enum_member, format_field_value, get_annotation_name, serialize_for_json
from dspy.signatures.signature import SignatureMeta
from dspy.signatures.utils import get_dspy_field_type

logger = logging.getLogger(__name__)


class FieldInfoWithName(NamedTuple):
    name: str
    info: FieldInfo


class JSONAdapter(Adapter):
    def __init__(self):
        pass

    def __call__(self, lm, lm_kwargs, signature, demos, inputs):
        inputs = self.format(signature, demos, inputs)
        inputs = dict(prompt=inputs) if isinstance(inputs, str) else dict(messages=inputs)

        try:
            provider = lm.model.split("/", 1)[0] or "openai"
            params = litellm.get_supported_openai_params(model=lm.model, custom_llm_provider=provider)
            if params and "response_format" in params:
                try:
                    response_format = _get_structured_outputs_response_format(signature)
                    outputs = lm(**inputs, **lm_kwargs, response_format=response_format)
                except Exception:
                    logger.debug(
                        "Failed to obtain response using signature-based structured outputs"
                        " response format: Falling back to default 'json_object' response format."
                        " Exception: {e}"
                    )
                    outputs = lm(**inputs, **lm_kwargs, response_format={"type": "json_object"})
            else:
                outputs = lm(**inputs, **lm_kwargs)

        except litellm.UnsupportedParamsError:
            outputs = lm(**inputs, **lm_kwargs)

        values = []

        for output in outputs:
            value = self.parse(signature, output)
            assert set(value.keys()) == set(
                signature.output_fields.keys()
            ), f"Expected {signature.output_fields.keys()} but got {value.keys()}"
            values.append(value)

        return values

    def format(self, signature, demos, inputs):
        messages = []

        # Extract demos where some of the output_fields are not filled in.
        incomplete_demos = [demo for demo in demos if not all(k in demo for k in signature.fields)]
        complete_demos = [demo for demo in demos if demo not in incomplete_demos]
        incomplete_demos = [
            demo
            for demo in incomplete_demos
            if any(k in demo for k in signature.input_fields) and any(k in demo for k in signature.output_fields)
        ]

        demos = incomplete_demos + complete_demos

        messages.append({"role": "system", "content": prepare_instructions(signature)})

        for demo in demos:
            messages.append(format_turn(signature, demo, role="user", incomplete=demo in incomplete_demos))
            messages.append(format_turn(signature, demo, role="assistant", incomplete=demo in incomplete_demos))

        messages.append(format_turn(signature, inputs, role="user"))

        return messages

    def parse(self, signature, completion):
        fields = json_repair.loads(completion)
        fields = {k: v for k, v in fields.items() if k in signature.output_fields}

        # attempt to cast each value to type signature.output_fields[k].annotation
        for k, v in fields.items():
            if k in signature.output_fields:
                fields[k] = parse_value(v, signature.output_fields[k].annotation)

        if fields.keys() != signature.output_fields.keys():
            raise ValueError(f"Expected {signature.output_fields.keys()} but got {fields.keys()}")

        return fields

    def format_turn(self, signature, values, role, incomplete=False):
        return format_turn(signature, values, role, incomplete)

    def format_fields(self, signature, values, role):
        fields_with_values = {
            FieldInfoWithName(name=field_name, info=field_info): values.get(
                field_name, "Not supplied for this particular example."
            )
            for field_name, field_info in signature.fields.items()
            if field_name in values
        }

        return format_fields(role=role, fields_with_values=fields_with_values)


def parse_value(value, annotation):
    if annotation is str:
        return str(value)

    parsed_value = value

    if isinstance(annotation, enum.EnumMeta):
        parsed_value = find_enum_member(annotation, value)
    elif isinstance(value, str):
        try:
            parsed_value = json.loads(value)
        except json.JSONDecodeError:
            try:
                parsed_value = ast.literal_eval(value)
            except (ValueError, SyntaxError):
                parsed_value = value

    return TypeAdapter(annotation).validate_python(parsed_value)


def _format_field_value(field_info: FieldInfo, value: Any) -> str:
    """
    Formats the value of the specified field according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself.

    Args:
      field_info: Information about the field, including its DSPy field type and annotation.
      value: The value of the field.

    Returns:
      The formatted value of the field, represented as a string.
    """
    if field_info.annotation is Image:
        raise NotImplementedError("Images are not yet supported in JSON mode.")

    return format_field_value(field_info=field_info, value=value, assume_text=True)


def format_fields(role: str, fields_with_values: Dict[FieldInfoWithName, Any]) -> str:
    """
    Formats the values of the specified fields according to the field's DSPy type (input or output),
    annotation (e.g. str, int, etc.), and the type of the value itself. Joins the formatted values
    into a single string, which is is a multiline string if there are multiple fields.

    Args:
      fields_with_values: A dictionary mapping information about a field to its corresponding
                          value.
    Returns:
      The joined formatted values of the fields, represented as a string.
    """

    if role == "assistant":
        d = fields_with_values.items()
        d = {k.name: v for k, v in d}
        return json.dumps(serialize_for_json(d), indent=2)

    output = []
    for field, field_value in fields_with_values.items():
        formatted_field_value = _format_field_value(field_info=field.info, value=field_value)
        output.append(f"[[ ## {field.name} ## ]]\n{formatted_field_value}")

    return "\n\n".join(output).strip()


def format_turn(signature: SignatureMeta, values: Dict[str, Any], role, incomplete=False) -> Dict[str, str]:
    """
    Constructs a new message ("turn") to append to a chat thread. The message is carefully formatted
    so that it can instruct an LLM to generate responses conforming to the specified DSPy signature.

    Args:
        signature: The DSPy signature to which future LLM responses should conform.
        values: A dictionary mapping field names (from the DSPy signature) to corresponding values
            that should be included in the message.
        role: The role of the message, which can be either "user" or "assistant".
        incomplete: If True, indicates that output field values are present in the set of specified
            ``values``. If False, indicates that ``values`` only contains input field values.
    Returns:
        A chat message that can be appended to a chat thread. The message contains two string fields:
        ``role`` ("user" or "assistant") and ``content`` (the message text).
    """
    content = []

    if role == "user":
        fields: Dict[str, FieldInfo] = signature.input_fields
        if incomplete:
            content.append("This is an example of the task, though some input or output fields are not supplied.")
    else:
        fields: Dict[str, FieldInfo] = signature.output_fields

    if not incomplete:
        field_names: KeysView = fields.keys()
        if not set(values).issuperset(set(field_names)):
            raise ValueError(f"Expected {field_names} but got {values.keys()}")

    formatted_fields = format_fields(
        role=role,
        fields_with_values={
            FieldInfoWithName(name=field_name, info=field_info): values.get(
                field_name, "Not supplied for this particular example."
            )
            for field_name, field_info in fields.items()
        },
    )
    content.append(formatted_fields)

    if role == "user":

        def type_info(v):
            return (
                f" (must be formatted as a valid Python {get_annotation_name(v.annotation)})"
                if v.annotation is not str
                else ""
            )

        # TODO: Consider if not incomplete:
        content.append(
            "Respond with a JSON object in the following order of fields: "
            + ", then ".join(f"`{f}`{type_info(v)}" for f, v in signature.output_fields.items())
            + "."
        )

    return {"role": role, "content": "\n\n".join(content).strip()}


def enumerate_fields(fields):
    parts = []
    for idx, (k, v) in enumerate(fields.items()):
        parts.append(f"{idx+1}. `{k}`")
        parts[-1] += f" ({get_annotation_name(v.annotation)})"
        parts[-1] += f": {v.json_schema_extra['desc']}" if v.json_schema_extra["desc"] != f"${{{k}}}" else ""

    return "\n".join(parts).strip()


def prepare_instructions(signature: SignatureMeta):
    parts = []
    parts.append("Your input fields are:\n" + enumerate_fields(signature.input_fields))
    parts.append("Your output fields are:\n" + enumerate_fields(signature.output_fields))
    parts.append("All interactions will be structured in the following way, with the appropriate values filled in.")

    def field_metadata(field_name, field_info):
        type_ = field_info.annotation

        if get_dspy_field_type(field_info) == "input" or type_ is str:
            desc = ""
        elif type_ is bool:
            desc = "must be True or False"
        elif type_ in (int, float):
            desc = f"must be a single {type_.__name__} value"
        elif inspect.isclass(type_) and issubclass(type_, enum.Enum):
            desc = f"must be one of: {'; '.join(type_.__members__)}"
        elif hasattr(type_, "__origin__") and type_.__origin__ is Literal:
            desc = f"must be one of: {'; '.join([str(x) for x in type_.__args__])}"
        else:
            desc = "must be pareseable according to the following JSON schema: "
            desc += json.dumps(pydantic.TypeAdapter(type_).json_schema())

        desc = (" " * 8) + f"# note: the value you produce {desc}" if desc else ""
        return f"{{{field_name}}}{desc}"

    def format_signature_fields_for_instructions(role, fields: Dict[str, FieldInfo]):
        return format_fields(
            role=role,
            fields_with_values={
                FieldInfoWithName(name=field_name, info=field_info): field_metadata(field_name, field_info)
                for field_name, field_info in fields.items()
            },
        )

    parts.append("Inputs will have the following structure:")
    parts.append(format_signature_fields_for_instructions("user", signature.input_fields))
    parts.append("Outputs will be a JSON object with the following fields.")
    parts.append(format_signature_fields_for_instructions("assistant", signature.output_fields))
    # parts.append(format_fields({BuiltInCompletedOutputFieldInfo: ""}))

    instructions = textwrap.dedent(signature.instructions)
    objective = ("\n" + " " * 8).join([""] + instructions.splitlines())
    parts.append(f"In adhering to this structure, your objective is: {objective}")

    # parts.append("You will receive some input fields in each interaction. " +
    #              "Respond only with the corresponding output fields, starting with the field " +
    #              ", then ".join(f"`{f}`" for f in signature.output_fields) +
    #              ", and then ending with the marker for `completed`.")

    return "\n\n".join(parts).strip()


def _get_structured_outputs_response_format(signature: SignatureMeta) -> pydantic.BaseModel:
    """
    Obtains the LiteLLM / OpenAI `response_format` parameter for generating structured outputs from
    an LM request, based on the output fields of the specified DSPy signature.

    Args:
        signature: The DSPy signature for which to obtain the `response_format` request parameter.
    Returns:
        A Pydantic model representing the `response_format` parameter for the LM request.
    """

    def filter_json_schema_extra(field_name: str, field_info: FieldInfo) -> FieldInfo:
        """
        Recursively filter the `json_schema_extra` of a FieldInfo to exclude DSPy internal attributes
        (e.g. `__dspy_field_type`) and remove descriptions that are placeholders for the field name.
        """
        field_copy = deepcopy(field_info)  # Make a copy to avoid mutating the original

        # Update `json_schema_extra` for the copied field
        if field_copy.json_schema_extra:
            field_copy.json_schema_extra = {
                key: value
                for key, value in field_info.json_schema_extra.items()
                if key not in ("desc", "__dspy_field_type")
            }
            field_desc = field_info.json_schema_extra.get("desc")
            if field_desc is not None and field_desc != f"${{{field_name}}}":
                field_copy.json_schema_extra["desc"] = field_desc

        # Handle nested fields
        if hasattr(field_copy.annotation, "__pydantic_model__"):
            # Recursively update fields of the nested model
            nested_model = field_copy.annotation.__pydantic_model__
            updated_fields = {
                key: filter_json_schema_extra(key, value) for key, value in nested_model.__fields__.items()
            }
            # Create a new model with the same name and updated fields
            field_copy.annotation = create_model(nested_model.__name__, **updated_fields)

        return field_copy

    output_pydantic_fields = {
        key: (value.annotation, filter_json_schema_extra(key, value)) for key, value in signature.output_fields.items()
    }
    return create_model("DSPyProgramOutputs", **output_pydantic_fields)



from setuptools import find_packages, setup
import os

# Read the content of the README file
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

# Read the content of the requirements.txt file
with open("requirements.txt", encoding="utf-8") as f:
    requirements = f.read().splitlines()

metadata = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "dspy", "__metadata__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), metadata)
f.close()


setup(	
    name=metadata["__name__"],
    version=metadata["__version__"],
    description=metadata["__description__"],
    url=metadata["__url__"],
    author=metadata["__author__"],
    author_email=metadata["__author_email__"],
    long_description=long_description,	
    long_description_content_type="text/markdown",	
    license="MIT License",	
    packages=find_packages(include=["dspy.*", "dspy"]),	
    python_requires=">=3.9",
    install_requires=requirements,	

    extras_require={
        "chromadb": ["chromadb~=0.4.14"],
        "lancedb": ["lancedb~=0.11.0"],
        "qdrant": ["qdrant-client", "fastembed"],
        "marqo": ["marqo~=3.1.0"],
        "mongodb": ["pymongo~=3.12.0"],
        "pinecone": ["pinecone-client~=2.2.4"],
        "weaviate": ["weaviate-client~=4.6.5"],
        "faiss-cpu": ["sentence_transformers", "faiss-cpu"],
        "milvus": ["pymilvus~=2.3.7"],
        "google-vertex-ai": ["google-cloud-aiplatform==1.43.0"],
        "myscale":["clickhouse-connect"],
        "snowflake": ["snowflake-snowpark-python"],
        "fastembed": ["fastembed"],
        "groq": ["groq~=0.8.0"],
        "langfuse": ["langfuse~=2.36.1"],
        "pgvector": ["psycopg2~=2.9.9","pgvector~=0.2.5"],
        "falkordb": ["falkordb", "redis", "async-timeout"]
    },
    classifiers=[	
        "Development Status :: 3 - Alpha",	
        "Intended Audience :: Science/Research",	
        "License :: OSI Approved :: MIT License",	
        "Operating System :: POSIX :: Linux",	
        "Programming Language :: Python :: 3",	
        "Programming Language :: Python :: 3.8",	
        "Programming Language :: Python :: 3.9",	
    ],	
)	



import logging
import sys
from pathlib import Path

import cloudpickle
import importlib_metadata
import ujson

logger = logging.getLogger(__name__)


def get_dependency_versions():
    cloudpickle_version = '.'.join(cloudpickle.__version__.split('.')[:2])
    return {
        "python": f"{sys.version_info.major}.{sys.version_info.minor}",
        "dspy": importlib_metadata.version("dspy"),
        "cloudpickle": cloudpickle_version,
    }


def load(path):
    """Load saved DSPy model.

    This method is used to load a saved DSPy model with `save_program=True`, i.e., the model is saved with cloudpickle.

    Args:
        path (str): Path to the saved model.

    Returns:
        The loaded model, a `dspy.Module` instance.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"The path '{path}' does not exist.")

    with open(path / "metadata.json", "r") as f:
        metadata = ujson.load(f)

    dependency_versions = get_dependency_versions()
    saved_dependency_versions = metadata["dependency_versions"]
    for key, saved_version in saved_dependency_versions.items():
        if dependency_versions[key] != saved_version:
            logger.warning(
                f"There is a mismatch of {key} version between saved model and current environment. You saved with "
                f"`{key}=={saved_version}`, but now you have `{key}=={dependency_versions[key]}`. This might cause "
                "errors or performance downgrade on the loaded model, please consider loading the model in the same "
                "environment as the saving environment."
            )

    with open(path / "program.pkl", "rb") as f:
        return cloudpickle.load(f)



from dspy.clients.lm import LM
from dspy.clients.provider import Provider, TrainingJob
from dspy.clients.base_lm import BaseLM, inspect_history
from dspy.clients.embedding import Embedder
import litellm
import os
from pathlib import Path
from litellm.caching import Cache

DISK_CACHE_DIR = os.environ.get("DSPY_CACHEDIR") or os.path.join(Path.home(), ".dspy_cache")
DISK_CACHE_LIMIT = int(os.environ.get("DSPY_CACHE_LIMIT", 3e10))  # 30 GB default

# TODO: There's probably value in getting litellm to support FanoutCache and to separate the limit for
# the LM cache from the embeddings cache. Then we can lower the default 30GB limit.
litellm.cache = Cache(disk_cache_dir=DISK_CACHE_DIR, type="disk")

if litellm.cache.cache.disk_cache.size_limit != DISK_CACHE_LIMIT:
    litellm.cache.cache.disk_cache.reset("size_limit", DISK_CACHE_LIMIT)

litellm.telemetry = False

# Turn off by default to avoid LiteLLM logging during every LM call.
litellm.suppress_debug_info = True

if "LITELLM_LOCAL_MODEL_COST_MAP" not in os.environ:
    # Accessed at run time by litellm; i.e., fine to keep after import
    os.environ["LITELLM_LOCAL_MODEL_COST_MAP"] = "True"


def enable_litellm_logging():
    litellm.suppress_debug_info = False


def disable_litellm_logging():
    litellm.suppress_debug_info = True


__all__ = [
    "BaseLM",
    "LM",
    "Provider",
    "TrainingJob",
    "inspect_history",
    "Embedder",
    "enable_litellm_logging",
    "disable_litellm_logging",
]



import textwrap
from dataclasses import dataclass
from typing import Generic, Literal, TypeVar

import pydantic

import dspy
from dspy import BaseModel
from dspy.functional.functional import TypedChainOfThought, TypedPredictor
from dspy.signatures import Signature
from dspy.signatures.field import InputField, OutputField

# TODO:
# - Parallelize the generation of new signatures when we have multiple predictors
# - Consider generating multiple new signatures at once, which we can test in parallel
# - Consider using the prompt optimizer to optimize the prompt optimizer :O


def make_info(signature: type[Signature]) -> BaseModel:
    """Creates a SignatureInfo pydantic type, that describes the Signature.

    Returns an instance of this type, with the instructions and field descriptions of the input type.
    """
    # First, create the SignatureInfo type
    fields = {
        "instructions": (str, pydantic.Field(description="The instructions for the task")),
    }
    for name in signature.fields:
        fields[name + "_prefix"] = (str, pydantic.Field(description=f"The prefix for {name}"))
        fields[name + "_desc"] = (str, pydantic.Field(description=f"The description for {name}"))
    SignatureInfo = pydantic.create_model(  # noqa: N806
        f"SignatureInfo[{signature.__name__}]",
        **fields,
    )

    # Add a method to convert the SignatureInfo back into a Signature
    def to_signature(info):
        new_signature = signature.with_instructions(info.instructions)
        for name in signature.fields:
            new_signature = new_signature.with_updated_fields(
                name,
                prefix=getattr(info, name + "_prefix"),
                desc=getattr(info, name + "_desc"),
            )
        return new_signature

    SignatureInfo.to_signature = to_signature

    # Finally, make an instance of the SignatureInfo type with the signature's
    # default instructions and field descriptions
    values = {"instructions": signature.instructions}
    for name, field in signature.fields.items():
        values[name + "_prefix"] = field.json_schema_extra["prefix"]
        values[name + "_desc"] = field.json_schema_extra["desc"]
    return SignatureInfo(**values)


T = TypeVar("T", bound=BaseModel)


# Note: This function wouldn't be necessary if we could make the number of prompts a generic parameter of the class,
# but alas it seems like this isn't possible in Python right now. The main reason being that types and generics only
# live inside the type system, and can't be used to generate code at runtime.
def make_initial_signature(n_prompts: int) -> type[Signature]:
    """Creates a GenerateInstructionInitial signature with the given number of initial prompts."""

    class GenerateInstructionInitial(Signature, Generic[T]):
        # TODO: Can we make textwrap default/automatic in all signatures?
        __doc__ = textwrap.dedent(
            """\
        You are a creative instruction optimizer for large language models.

        I will give you a ``signature`` of fields (inputs and outputs) in English.
        Your task is to propose variations of the signature that will lead a good language model.

        Be very creative and think out of the box.
        You can use as long instructions as you want.
        Consider using inspiration such as:
        Openers:
        - You are as smart as ChatGPT.
        - You are highly intelligent.
        - You are an expert mathematician.
        - You are a professor of mathematics.
        Task Descriptions:
        - Be concise in your answer.
        - Be as clear as possible.
        - Use lots of creativity.
        Closers:
        - This will be fun!
        - Take a deep breath and think carefully.
        - I really need your help!
        """
        )

        basic_signature: T = InputField()
        proposed_signatures: list[T] = OutputField(
            desc=f"A list of {n_prompts} very different variations of the basic signature",
            min_items=n_prompts,
            max_items=n_prompts,
        )

    return GenerateInstructionInitial


def generate_with_avoidance(signatures_to_avoid: list[BaseModel]) -> type[Signature]:
    class GenerateSignature(dspy.Signature, Generic[T]):
        __doc__ = textwrap.dedent(
            """\
        You are an instruction optimizer for large language models.

        I will give some task instructions I've tried, along with their corresponding validation scores.
        - The instructions are arranged in order based on their scores, where higher scores indicate better quality.
        - Your task is to propose a new instruction that will lead a good language model to perform the task even better.
        - Be creative, and think out of the box.
        - Don't repeat instructions, descriptions and prefixes that have already been attempted.
        """
        )

        analysis: str = OutputField(desc="Consider what made the previous instructions good or bad.")
        proposed_signature: T = OutputField(desc="A signature that will likely lead to a high score.")
        score: float = OutputField(
            desc="The expected score for the new signature. Don't write anything after this number.",
        )

        @pydantic.field_validator("proposed_signature")
        @classmethod
        def check_signature_not_attempted(cls, s: T) -> T:
            if s in signatures_to_avoid:
                raise ValueError("Never propose a signature already in the list above.")
            return s

    return GenerateSignature


@dataclass
class OptimizerResult:
    program: dspy.Program
    signatures: list[dict[str, Signature]]
    scores: list[float]


def optimize_signature(
    student,
    evaluator,
    n_iterations=10,
    sorted_order: Literal["increasing", "decreasing", "chronological"] = "increasing",
    strategy: Literal["last", "best"] = "best",
    max_examples=20,
    # Formerly part of the constructor
    prompt_model=None,
    initial_prompts=2,
    verbose=False,
    max_retries: int = 3,
) -> dspy.Program:
    """Create a new program that is optimized for the given task.

    `student` is a program that needs to be optimized,
    note that it may be zero-shot or already pre-optimized for demos != [].

    Parameters
    ----------
    student : dspy.Program
        The program to optimize.
    evaluator : dspy.Evaluator
        The evaluator to use to score the program.
    n_iterations : int, optional
        The number of iterations to run, by default 10
    max_examples : int, optional
        The maximum number of examples to use for the evaluator, by default 20
    sorted_order : Literal["increasing", "decreasing", "chronological"], optional
        The order in which to sort the scores, by default "increasing"
    strategy : Literal["last", "best"], optional
        The strategy to use to select the final program, by default "best"
    prompt_model : dspy.LanguageModel, optional
        The language model to use to generate prompts, by default None
    initial_prompts : int, optional
        The number of initial prompts to generate, by default 2.
        Note that we also use the "plain" signature as a prompt, so the total number of prompts is initial_prompts + 1.
    verbose : bool, optional
        Whether to print debug information, by default False
    max_retries : int
        The number of retries to use when generating new signatures, by default 3.

    Notes:
    -----
    We don't support temperatures, since it tends to break the typed generation.
    """
    if n_iterations < 1 + initial_prompts:
        raise ValueError("n_iterations must be at least 1 + initial_prompts")

    prompt_model = prompt_model or dspy.settings.lm
    MyGenerateInstructionInitial = make_initial_signature(initial_prompts)  # noqa: N806

    module = student.deepcopy()
    # In contrast to the original implementation, we don't want the Predict's, but the TypedPredictor's.
    # This is because TypedPredictor changes the signature before it runs forward. So changing the signature
    # on the Predicts won't work.
    named_predictors = [
        (name, module)
        for name, module in module.named_sub_modules()
        if isinstance(module, TypedPredictor) and not getattr(module, "_compiled", False)
    ]
    if not named_predictors:
        raise ValueError("No unfrozen/uncompiled TypedPredictors found in the module.")
    if verbose:
        print(f"Found {len(named_predictors)} typed predictors to optimize.")

    candidates = {}
    scores = []

    # First round, just use initial prompts
    for name, p in named_predictors:
        candidates[name] = [make_info(p.signature)]

    # Make some initial candidates
    with dspy.settings.context(lm=prompt_model):
        # TODO: Parallelize this
        for name, _p in named_predictors:
            if verbose:
                print(f"Generating {initial_prompts} initial signatures for {name}...")
            info = candidates[name][0]  # Use initial info, to make sure types are identical
            generator = TypedChainOfThought(MyGenerateInstructionInitial[type(info)], max_retries=max_retries)
            candidates[name] += generator(
                basic_signature=info,
            ).proposed_signatures
            assert len(candidates[name]) == initial_prompts + 1  # Basic signature + initial prompts

    # Main loop of scoring + generating new candidates
    for i in range(n_iterations):
        if verbose:
            print("\n" + "=" * 80)
            print(f"Running eval iteration {i}...")

        # Install signatures
        for name, p in named_predictors:
            p.signature = candidates[name][i].to_signature()

        # Run evaluator given by user
        score = evaluator(module)
        if isinstance(score, tuple):
            scores.append(score[0])
        else:
            scores.append(score)

        # If we are still testing initial prompts, continue
        if i + 1 < len(next(iter(candidates.values()))):
            continue

        # If we are done, there's no need to generate new candidates
        if i + 1 == n_iterations:
            break

        # Otherwise generate the next candidate
        with dspy.settings.context(lm=prompt_model):
            # TODO: Parallelize this
            for name, _p in named_predictors:
                SignatureInfo = type(candidates[name][0])  # noqa: N806

                demos = [dspy.Example(proposed_signature=info, score=sc) for info, sc in zip(candidates[name], scores)]
                if sorted_order == "chronological":
                    demos = demos[-max_examples:]
                elif sorted_order == "increasing":
                    demos.sort(key=(lambda x: x.score), reverse=False)
                    demos = demos[-max_examples:]
                elif sorted_order == "decreasing":
                    demos.sort(key=(lambda x: x.score), reverse=True)
                    demos = demos[:max_examples]
                else:
                    raise ValueError(f"Invalid sorted_order: {sorted_order}")

                # We can only tell the LM to avoid the signatures we are actually giving it as demos.
                avoid = [ex.proposed_signature for ex in demos]
                generator = TypedPredictor(generate_with_avoidance(avoid)[SignatureInfo], max_retries=max_retries)
                generator.predictor.demos = demos

                if verbose:
                    print(f"Generating new signature for {name}...")
                new_signature = generator().proposed_signature
                candidates[name].append(new_signature)

    if strategy == "last":
        pass
    elif strategy == "best":
        i = scores.index(max(scores))
        if verbose:
            print(f"Best signature: {i} with score: {scores[i]}")
        for name, p in named_predictors:
            p.signature = candidates[name][i].to_signature()
    else:
        raise ValueError(f"Invalid strategy: {strategy}")

    return OptimizerResult(
        program=module,
        signatures=[{name: sigs[i].to_signature() for name, sigs in candidates.items()} for i in range(n_iterations)],
        scores=scores,
    )



from dspy.adapters.base import Adapter
from dspy.adapters.chat_adapter import ChatAdapter
from dspy.adapters.json_adapter import JSONAdapter
from dspy.adapters.image_utils import Image

__all__ = [
    "Adapter",
    "ChatAdapter",
    "JSONAdapter",
    "Image",
]
